import itertools
from collections import deque
from scipy.spatial.distance import euclidean
import numpy as np
import rdkit.Chem.AllChem as rdkit

from ..base import Topology
from ....utilities import (centroid,
                           vector_theta,
                           add_fragment_props,
                           normalize_vector)


class Vertex:
    """
    Used to represent the vertices of cage polyhedra.

    This class stores information about the vertices which make up a
    cage's structure.

    Attributes
    ----------
    coord : :class:`numpy.array`
        An array which holds the x, y and z coordinates of the
        vertex, in that order.

    connected : :class:`list` of :class:`Edge` or :class:`Vertex`
        This :class:`list` holds the :class:`Edge` or :class:`Vertex`
        instances which represent the edges or vertices connected to
        this one.

    fg_ids : :class:`list` of :class:`int`
        This :class:`list` holds the ids of fgs which belong to the
        building block placed on the vertex and will form bonds. The
        ids reflect those in the macromolecule not in the building
        block itself.

    fg_position_pairs : :class:`list`
        A :class:`list` of the form

        .. code-block:: python

            fg_position_pairs = [(4, v1), (8, v2)]

        Where ``v1`` and ``v2`` are :class:`Vertex` instances.

        Each fg which  is paired to a neighboring vertex or
        edge first. Only after this is the fg-fg pairing
        performed. The fg-edge/vertex pairing is stored here. The
        :class:`int` represents the id of the fg.

    distances : :class:`list`
        A :class:`list` of the form

        .. code-block:: python

            distances = [(15.2, 5, 10), (18.4, 3, 8), ...]

        After fgs have been associated with vertices to which
        they join, the idividual fgs are paired up. To do this the
        distance between every fg on the paired vertex and the
        fg which is paired to the vertex is found. This
        information is stored here the first element of each
        :class:`tuple` the distance, and the second and third elements
        represent the fgs.

    id_ : :class:`object`, optional
        An id to identify the vertex. Used by
        :meth:`CageTopology.place_mols`.

    custom_position : :class:`bool`
        A flag to inidcate if the position of :attr:`coord` was derived
        from connected vertices or set manually. If the position is
        derived from the vertices then the position of the building
        block being placed is derived from the bonder atoms in the
        conncected vertices.

    """

    def __init__(self, x, y, z, id_=None, custom_position=True):
        """
        Initializes a :class:`Vertex`.

        Parameters
        ----------
        x : :class:`int` or :class:`float`
            The x coordinate.

        y : :class:`int` or :class:`float`
            The y coordiante

        z : :class:`int` or :class:`float`
            The z coordiante

        id : :class:`object`, optional
            An id to identify the vertex.

        custom_position : :class:`bool`, optional
            See :attr:`custom_position`.

        """

        self.coord = np.array([x, y, z])
        self.custom_position = custom_position
        self.connected = []
        self.fg_ids = []
        self.fg_position_pairs = []
        self.distances = []
        self.id = id_

    @classmethod
    def vertex_init(cls, *vertices):
        """
        Intializes the :class:`Vertex` from other vertices.

        This initalizer automatically calculates the position of the
        vertex from the positions of the vertices provided to the
        initializer. Its position is set to the centroid of the
        provided vertices.

        The :attr:`connected` attributes of all the involved vertices
        are also updated.

        Parameters
        ----------
        vertices : :class:`tuple` of :class:`Vertex`
            Holds :class:`Vertex` objects used for initialization.

        Returns
        -------
        :class:`Vertex`
            The initialized vertex.

        """

        # Get the position of this Vertex, as the centroid of the
        # supplied vertices.
        obj = cls(*centroid(*(v.coord for v in vertices)),
                  custom_position=False)
        # Update the `connected` attributes.
        obj.connected.extend(vertices)
        for v in vertices:
            v.connected.append(obj)
        return obj

    def place_mol(self,
                  scale,
                  building_block,
                  aligner=0,
                  aligner_edge=0,
                  macro_mol=None):
        """
        Place a :class:`.StructUnit3` building block on the vertex.

        The orientation of the building block is aligned with 2
        parameters. Firstly, the normal of the plane of fgs of
        the building block is aligned with the normal of the plane
        formed by the edges connected to the vertex. Because the normal
        of the plane of fgs always points in the direction of
        the building block's centroid, this alignment causes the bulk
        of the building block  molecule to point away from the center
        of the cage.

        Secondly, the building block is rotated so that a fg
        is aligned perfectly with an edge. This reduces the rms
        distance between the edges and fgs to some extent,
        probably.

        Parameters
        ----------
        scale : :class:`float`
            The amount by which the size of the topology is scaled.

        building_block : :class:`.StructUnit3`
            The building block molecule to be placed on a vertex.

        aligner : :class:`int`, optional
            The ``fg_id`` of the aligned fg.

        aligner_edge : :class:`int`, optional
            The index of an edge in :attr:`connected`. It is the edge
            with which `aligner` is aligned.

        macro_mol : :class:`.MacroMolecule`, optional
            The macromolecule being built. Used for vertex only cage
            topologies where the position of some of the vertices
            is derived from the positions of the fgs on
            connected vertices.

        Returns
        -------
        :class:`rdkit.Chem.rdchem.Mol`
            The ``rdkit`` instance holding the building block molecule
            with the coordinates placed on the vertex and orientation
            set as described in the docstring.

        """

        icoord = building_block.position_matrix()
        # Flush the list of data from previous molecules.
        self.distances = []

        # The method first aligns the normal of the fg plane
        # to the normal of the edge plane. This means the bulk of the
        # building block is always pointed away from the center of the
        # molecule.
        building_block.set_orientation2(self.edge_plane_normal(scale))

        # Next, define the direction vector going from the edge
        # centroid to the edge with which the fg is aligned.

        building_block.set_bonder_centroid(
                            self.bonder_centroid(macro_mol, scale))
        vector = (self.connected[aligner_edge].coord*scale -
                  self.edge_centroid(scale))

        # Minimize the angle between these things by rotating about the
        # normal of the edge plane.
        building_block.minimize_theta2(aligner,
                                       vector,
                                       self.edge_plane_normal(scale))

        mol = rdkit.Mol(building_block.mol)
        building_block.set_position_from_matrix(icoord)
        return mol

    def edge_plane_normal(self, scale):
        """
        Return the normal of the plane formed by the connected edges.

        The normal is set such that it always points away from the
        origin.

        Parameters
        ----------
        scale : :class:`float`
            The amount by which the size of the topology is scaled.

        Returns
        -------
        :class:`numpy.array`
            A normalized vector which defines the normal pointed away
            from the origin.

        """
        # Get two of the direction vectors running between the edges.
        v1, v2 = itertools.islice(self.edge_direction_vectors(scale), 2)
        # To get the normal to the plane get the cross product of these
        # vectors. Normalize it.
        normal = normalize_vector(np.cross(v1, v2))

        # To check that the normal is pointing away from the center of
        # cage, find the angle, `theta`, between it and one of the
        # position vectors of the edges on the plane. Assuming that the
        # center of the cage is at the origin, which it should be as
        # this is specified in the documentation, if the angle between
        # the normal the position vector is less than 90 degrees they
        # point in the same general direction. If the angle is greater
        # than 90 degrees it means that they are pointing in opposite
        # directions. If this is the case make sure to multiply the
        # nomral by -1 in all axes so that it points in the correct
        # direction while still acting as the normal to the plane.
        theta = vector_theta(normal, self.connected[0].coord*scale)

        if theta > np.pi/2:
            normal *= -1

        return normal

    def edge_plane(self, scale):
        """
        Return coefficients of plane of edges connected to the vertex.

        A plane is defined by the scalar plane equation::

            ax + by + cz = d.

        This method returns the ``a``, ``b``, ``c`` and ``d``
        coefficients of this equation for the plane formed by the
        connected edges. The coefficents ``a``, ``b`` and ``c`` decribe
        the normal vector to the plane. The coefficent ``d`` is found
        by substituting these coefficients along with the ``x``, ``y``
        and ``z`` variables in the scalar equation and solving for
        ``d``. The variables ``x``, ``y`` and ``z`` are substituted by
        the coordinate of some point on the plane. For example, the
        position of one of the connected edges.

        Parameters
        ----------
        scale : :class:`float`
            The amount by which the size of the topology is scaled.

        Returns
        -------
        :class:`numpy.array`
            This array has the form ``[a, b, c, d]`` and represents the
            scalar equation of the plane formed by the connected edges.

        References
        ----------
        https://tinyurl.com/okpqv6

        """

        edge_coord = self.edges[0].coord*scale
        d = -np.sum(self.edge_plane_normal(scale) * edge_coord)
        return np.append(self.edge_plane_normal(scale), d)

    def edge_direction_vectors(self, scale):
        """
        Yields direction vectors between edges connected to the vertex.

        Parameters
        ---------
        scale : :class:`float`
            The amount by which the size of the topology is scaled.

        Yields
        ------
        :class:`numpy.array`
            A normalized direction vector running from one edge
            connected to the vertex to another.

        """

        for edge1, edge2 in itertools.combinations(self.connected, 2):
            yield normalize_vector(scale*(edge1.coord-edge2.coord))

    def edge_coord_matrix(self, scale):
        """
        Return matrix holding coords of edges joined to the vertex.

        Parameters
        ----------
        scale : :class:`float`
            The amount by which the size of the topology is scaled.

        Returns
        -------
        :class:`numpy.matrix`
            The matrix of shape ``[n, 3]``, where ``n`` is the number
            of edges connected to the vertex. The row holds the x, y
            and z coordinates, respectively.

        """

        coords = []
        for edge in self.connected:
            coords.append(edge.coord*scale)
        return np.matrix(coords)

    def edge_centroid(self, scale):
        """
        Returns the centroid of the edges connected to the vertex.

        Parameters
        ----------
        scale : :class:`float`
            The amount by which the size of the topology is scaled.

        Returns
        -------
        :class:`numpy.array`
            An array which holds the x, y and z positions of the
            centroid of the edges connected to the vertex.

        """

        # The connected edges are held in the `edges`. To get the
        # centroid, add up all the x, y and z coordinates (separately)
        # and divide each sum by the number of edges.
        return (sum(edge.coord*scale for edge in self.connected) /
                len(self.connected))

    def bonder_centroid(self, macro_mol, scale):
        """
        Calculates the centroid of the fgs on the vertex.

        However if :attr:`custom_position` is ``True`` or
        `macro_mol` is ``None``, then :attr:`coord` is returned.

        Parameters
        ----------
        macro_mol : :class:`.MacroMolecule`
            The macromolecule being built.

        scale : :class:`float`
            The amount by which the size of the topology is scaled.

        Returns
        -------
        :class:`numpy.array`
            The centroid of the bonder atoms.

        """

        if self.custom_position or macro_mol is None:
            return self.coord*scale

        centroid = np.zeros((3, ))
        count = 0
        for v in self.connected:
            for fg, edge in v.fg_position_pairs:
                if edge is self:
                    centroid += macro_mol.fg_centroid(fg)
                    count += 1
        return centroid / count

    def __repr__(self):
        return "Vertex({:.3}, {:.3}, {:.3})".format(*self.coord)

    def __str__(self):
        return repr(self)


class Edge(Vertex):
    """
    Used to represent the edges of Cage polyhedra.

    This class stores information about the edges which make up a
    cage's structure.

    Attributes
    ----------
    direction : :class:`numpy.array`
        This vector represents the orientation of the edge. It is a
        normalized direction vector which runs from `v2` to `v1`,
        provided in the initalizer.

    """

    def __init__(self, v1, v2, id_=None, custom_position=False):
        """
        Initializes an :class:`Edge` instance.

        Parameters
        ----------
        v1 : :class:`Vertex`
            A vertex joined to the edge.

        v2 : :class:`Vertex`
            Another vertex joined to the edge.

        id_ : :class`str`, optional
            An id for the edge.

        custom_position : :class:`bool`
            See :attr:`Vertex.custom_position`


        """

        Vertex.__init__(self,
                        *centroid(v1.coord, v2.coord),
                        id_,
                        custom_position)

        self.connected.extend([v1, v2])
        v1.connected.append(self)
        v2.connected.append(self)

    def direction(self, macro_mol, scale):
        if self.custom_position or macro_mol is None:
            v1, v2 = self.connected
            return normalize_vector(v1.coord - v2.coord)

        fgs = []
        for v in self.connected:
            for fg, edge in v.fg_position_pairs:
                if edge is self:
                    fgs.append(macro_mol.fg_centroid(fg))
        return normalize_vector(fgs[0] - fgs[1])

    def place_mol(self, scale, linker, alignment, macro_mol):
        """
        Places a linker molecule on the coordinates of an edge.

        It also orientates the linker so that a the fgs sit
        exactly on the edge and the bulk of the linker points away from
        the center of the cage.

        Parameters
        ----------
        scale : :class:`float`
            The amount by which the size of the topology is scaled.

        linker : :class:`.StructUnit2`
            The linker which is to be placed and orientated as
            described in the docstring.

        alignment : :class:`int`
            ``1`` for parallel alignment with :attr:`direction` and
            ``-1`` for anti-parallel alignment with :attr:`direction`.

        macro_mol : :class:`.MacroMolecule`
            The macromolecule being constructed.

        Returns
        -------
        :class:`rdkit.Chem.rdchem.Mol`
            The ``rdkit`` instance holding the linker molecule with the
            coordinates placed on the edge and orientation set as
            described in the docstring.

        """

        icoord = linker.position_matrix()

        # Flush the lists from data of previous molecules.
        self.distances = []

        # Align then place the linker.
        linker.set_orientation2(self.direction(macro_mol, scale)*alignment)
        linker.minimize_theta2(self.coord*scale,
                               self.direction(macro_mol, scale))
        linker.set_bonder_centroid(self.bonder_centroid(macro_mol, scale))

        mol = rdkit.Mol(linker.mol)
        linker.set_position_from_matrix(icoord)
        return mol

    def __repr__(self):
        v1, v2 = self.connected
        return f"Edge({v1}, {v2})"

    def __str__(self):
        return repr(self)


class CageTopology(Topology):
    """
    An abstract base class for cage topologies.

    Attributes
    ----------
    A_alignments : :class:`list` of :class:`int`
        The length of this :class:`list` must be equal to the number of
        building blocks in the cage. When cages are built one of the
        fgs of each building block is aligned with an edge
        during placement. The :class:`int` is the ``fg_id`` of the
        aligned functional group.

        If ``None`` then the functional group with ``fg_id`` of ``0``
        is always aligned.

        For example,

        .. code-block:: python

            A_alignments = [0, 2, 1, 2]

        In this case there must be 4 building blocks in the cage. The
        the first building block has its first (``fg_id = 0``) fg
        aligned. The 2nd building block has the 3rd fg (``fg_id = 2``)
        aligned. The 3rd building block has the 2nd fg (``fg_id = 1``)
        aligned. The 4th building block has the 3rd fg (``fg_id = 2``)
        aligned.

    B_alignments : :class:`list` of :class:`int`
        The length of this :class:`list` should be euqal to the number
        of linkers in the cage. The linkers of a cage can have either 2
        functional groups or 3 or more, depending on the topology.

        When the linkers have 2 functional groups, the :class:`list`
        should hold either ``1`` or ``-1``. The value indicates that
        the linker is aligned parallel or antiparallel with the edge
        its placed on. For example, in a tetrahedral topology,

        .. code-block:: python

            B_alignments = [-1, 1, 1, -1, 1, -1]

        It indicates that the first linker is aligned anti parallel and
        the second is aligned in parallel, and so on.

        If the linkers have 3 or more functional groups, the values
        in :attr:`B_alignments` have the same role as
        :attr:`A_alignments`. The only difference is that by default
        the second fg is aligned, rather than the first.

    edge_alignments : :class:`list` of :class:`int`
        The length of the :class:`list` is equal to the number of
        building blocks in the cage. Each element is an :class:`int`
        which holds the id of an edge. For example,

        .. code-block:: python

            edge_alignments = [1, 2, 3, 4]

        then the first building block is aligned with the edge with
        :attr:`Vertex.id_` of ``1``, the second building block is
        aligned with the edge with :attr:`Vertex.id_` ``2`` and so on.
        For this to work, the edges must have their :attr:`Vertex.id_`
        attributes defined.

    bb_assignments : :class:`dict`
        A :class:`dict` of the form

        .. code-block:: python

            bb_assignments = {
                0: [0, 1, 3],
                1: [2]
                2: [0, 4, 5],
                3: [1, 2, 3]
            }

        This means the building block at index ``0`` of
        :attr:`.MacroMolecule.building_blocks` sits on the vertices
        of indices ``0``, ``1``, and ``3``. The building block at
        index ``1`` of :attr:`.MacroMolecule.building_blocks` sits on
        verticex of index ``2``.

        The building block at index ``2`` in
        :attr:`.MacroMolecule.building_blocks` also sits on the vertex
        of index ``0``. However each cage is composed of two sets of
        building blocks, called building blocks and linkers. This means
        that the building block at index ``0`` is builing block while
        the one at index ``2`` is a linker, or vice versa. Which one
        is which is automatically deduced by looking at the number of
        functional groups.

        If ``None`` then building blocks are assigned at random.

    """

    def __init__(self,
                 A_alignments=None,
                 B_alignments=None,
                 edge_alignments=None,
                 bb_assignments=None):

        if A_alignments is None:
            A_alignments = np.zeros(len(self.positions_A))
        if B_alignments is None:
            B_alignments = np.ones(len(self.positions_B))
        if edge_alignments is None:
            edge_alignments = [None for x in
                               range(len(self.positions_A))]

        super().__init__()
        self.A_alignments = A_alignments
        self.B_alignments = B_alignments
        self.edge_alignments = edge_alignments
        self.bb_assignments = bb_assignments

    def _bb_maps(self, macro_mol):
        """

        """

        bb_fgs = max(len(bb.functional_group_atoms()) for
                     bb in macro_mol.building_blocks)
        bb_map, lk_map = {}, {}

        if self.bb_assignments is None:
            bbs = [bb for bb in macro_mol.building_blocks if
                   len(bb.functional_group_atoms()) == bb_fgs]
            lks = [bb for bb in macro_mol.building_blocks if
                   len(bb.functional_group_atoms()) != bb_fgs]

            for i in range(len(self.positions_A)):
                bb_map[i] = np.random.choice(bbs)
            for i in range(len(self.positions_B)):
                lk_map[i] = np.random.choice(lks)

        else:

            for bb_index, positions in self.bb_assignments.items():
                bb = macro_mol.building_blocks[bb_index]
                n_fgs = len(bb.functional_group_atoms())
                map_ = bb_map if n_fgs == bb_fgs else lk_map
                for position in positions:
                    map_[position] = bb

        return bb_map, lk_map

    def bonded_fgs(self, macro_mol):
        """
        Joins up the separate building blocks which form the molecule.

        Parameters
        ----------
        macro_mol : :class:`.MacroMolecule`
            The macromolecule being assembled.

        Yields
        ------
        :class:`tuple` of :class:`int`
            The ``fg_ids`` of functional groups to be bonded.

        """

        # This loop finds all the distances between a fg paired with
        # a postion and all other fgs at the paired position.
        for position in self.positions_A:
            for fg1, vertex in position.fg_position_pairs:
                # Get all the distances between the fg and the other
                # fgs on the vertex. Store this information
                # on the vertex.
                for fg2 in vertex.fg_ids:
                    distance = macro_mol.fg_distance(fg1, fg2)
                    position.distances.append((distance, fg1, fg2))

        # This loop creates bonds between fgs at two different
        # positions so that each fg only bonds once and so that the
        # total length of all bonds made is minimzed.
        paired = set()
        for position in self.positions_A:
            for _, fg1, fg2 in sorted(position.distances):
                if fg1 in paired or fg2 in paired:
                    continue

                yield fg1, fg2
                paired.add(fg1)
                paired.add(fg2)

    def pair_fgs_with_positions(self, scale, macro_mol, vertex):
        """
        Matches fgs with the closest building block position.

        After a building block is placed on a position, each fg
        which forms a bond must be paired with the location of the
        building block to which it bonds. This function matches fgs
        and positions so that each is only present in one pairing and
        so that the total distance of the pairings is minimized.

        This updates of :attr:`Vertex.fg_position_pairs` attribute of
        `vertex`.

        Parameters
        ----------
        scale : :class:`float`
            The amount by which the size of the topology is scaled.

        macro_mol : :class:`.MacroMolecule`
            The macromolecule being buit.

        vertex : :class:`Vertex`
            The position at which all the atoms being paired are
            located.

        Returns
        -------
        None : :class:`NoneType`

        """

        # This loop looks at each fg which forms a new bond and all
        # the positions (not fgs) to which it may end up bonding. It
        # finds the distances of all the options.
        distances = []
        for fg in vertex.fg_ids:
            fg_coord = macro_mol.fg_centroid(fg)
            for position in vertex.connected:
                distance = euclidean(fg_coord, position.coord*scale)
                distances.append((distance, fg, position))

        # Sort the pairings of fgs with potential bonding position,
        # smallest first.
        distances.sort(key=lambda x: x[0])

        # This loop looks at all the potential pairings of fgs to
        # positions. It pairs the shortest combinations of fgs and
        # positions, making sure that each fg and position is only
        # paired once. The pairings are saved to the `
        # fg_positions_pairs` attribute of the position on which all
        # the fgs are placed.
        paired_pos = set()
        paired_ids = set()
        vertex.fg_position_pairs = []
        for _, fg, pos in distances:
            if fg in paired_ids or pos in paired_pos:
                continue
            vertex.fg_position_pairs.append((fg, pos))
            paired_ids.add(fg)
            paired_pos.add(pos)

    def place_mols(self, macro_mol):
        """
        Places all building block molecules on correct coordinates.

        The building block molecules are placed in their appropriate
        positions based on the topology. It does not join them.

        Also updates :attr:`.MacroMolecule.bb_counter`.

        Parameters
        ----------
        macro_mol : :class:`.MacroMolecule`
            The macromolecule being built.

        Returns
        -------
        None : :class:`NoneType`

        """

        macro_mol.mol = rdkit.Mol()
        bb_map, lk_map = self._bb_maps(macro_mol)
        scale = max(bb.max_diameter()[0] for bb in macro_mol.building_blocks)

        # This loop places all building-blocks* on the points at
        # `positions_A`. It then pairs all fgs which form a new bond
        # with the positions to which they will be bonding. It also
        # counts the nubmer of building-blocks* which make up the
        # structure.
        for i, position in enumerate(self.positions_A):
            bb = bb_map[i]
            n_bb = len(bb.functional_group_atoms())
            # Position the molecule on the vertex.
            aligner_edge_id = self.edge_alignments[i]
            aligner_edge = next((position.connected.index(x) for x in
                                 position.connected if
                                 x.id == aligner_edge_id), 0)
            bb_mol = position.place_mol(scale,
                                        bb,
                                        int(self.A_alignments[i]),
                                        aligner_edge)
            add_fragment_props(bb_mol,
                               macro_mol.building_blocks.index(bb),
                               i)

            bb_mol = self.update_fg_id(macro_mol, bb_mol)
            macro_mol.mol = rdkit.CombineMols(macro_mol.mol, bb_mol)
            # Update the counter each time a building-block* is added.
            macro_mol.bb_counter.update([bb])

            # Get ids of fgs which form new bonds.
            fg_ids = deque(maxlen=n_bb)
            for atom in macro_mol.mol.GetAtoms():
                if (atom.HasProp('fg_id') and
                   atom.GetIntProp('fg_id') not in fg_ids):
                    fg_ids.append(atom.GetIntProp('fg_id'))

            # Save the ids of fgs which form new bonds and pair them
            # up with positions.
            position.fg_ids = sorted(fg_ids)
            self.pair_fgs_with_positions(scale, macro_mol, position)

        # This loop places all linkers on the points at `positions_B`.
        # It then saves all fgs which form a new bond to the position
        # they are found at. It also counts the number of linkers which
        # make up the structure.
        for i, position in enumerate(self.positions_B):
            lk = lk_map[i]
            n_lk = len(lk.functional_group_atoms())
            lk_mol = position.place_mol(scale,
                                        lk,
                                        int(self.B_alignments[i]),
                                        macro_mol=macro_mol)
            add_fragment_props(lk_mol,
                               macro_mol.building_blocks.index(lk),
                               i)
            lk_mol = self.update_fg_id(macro_mol, lk_mol)
            macro_mol.mol = rdkit.CombineMols(macro_mol.mol, lk_mol)
            # Update the counter each time a linker is added.
            macro_mol.bb_counter.update([lk])

            # Get ids of fgs which form new bonds.
            fg_ids = deque(maxlen=n_lk)
            for atom in macro_mol.mol.GetAtoms():
                if (atom.HasProp('fg_id') and
                   atom.GetIntProp('fg_id') not in fg_ids):
                    fg_ids.append(atom.GetIntProp('fg_id'))

            # Save the ids of fgs which form new bonds.
            position.fg_ids = list(fg_ids)


class VertexOnlyCageTopology(CageTopology):
    """
    Abstract base class for cage topologies using ditopics.

    """

    ...


class NoLinkerCageTopology(CageTopology):
    """
    Abstract base class for cage topologies without linkers.

    This means that all building blocks have the same number of
    functional groups.

    Attributes
    ----------
    alignments : :class:`list` of :class:`int`
        See :attr:`CageTopology.A_alignments`

    bb_assignments : :class:`list` of :class:`int`
        For each vertex, the :class:`int` is the index of a
        building block in :attr:`.MacroMolecule.building_blocks`.
        It is the building block to be placed on that vertex. Can be
        ``None`` if a random building block should be placed.

    """

    def __init__(self, alignments=None, bb_assignments=None):
        if alignments is None:
            alignments = np.zeros(len(self.positions_A))
        if bb_assignments is None:
            bb_assignments = [None for i in range(len(self.positions_A))]

        self.alignments = alignments
        self.bb_assignments = bb_assignments
        self.connect()
        self.react_del = True

    def place_mols(self, macro_mol):

        macro_mol.mol = rdkit.Mol()
        scale = max(bb.max_diameter()[0] for bb in macro_mol.building_blocks)

        for position, orientation, bb_index in zip(self.positions_A,
                                                   self.alignments,
                                                   self.bb_assignments):
            if bb_index is None:
                bb = np.random.choice(macro_mol.building_blocks)
            else:
                bb = macro_mol.building_blocks[bb_index]
            ipos = bb.position_matrix()
            n_bb = len(bb.functional_group_atoms())

            mol = position.place_mol(scale, bb, int(orientation))
            mol = self.update_fg_id(macro_mol, mol)
            macro_mol.mol = rdkit.CombineMols(macro_mol.mol, mol)
            macro_mol.bb_counter.update([bb])

            fg_ids = deque(maxlen=n_bb)
            for atom in macro_mol.mol.GetAtoms():
                if (atom.HasProp('fg_id') and
                   atom.GetIntProp('fg_id') not in fg_ids):
                    fg_ids.append(atom.GetIntProp('fg_id'))

            position.fg_ids = sorted(fg_ids)
            self.pair_fgs_with_positions(scale, macro_mol, position)
            bb.set_position_from_matrix(ipos)

    @classmethod
    def connect(cls):
        """
        Updates each :attr:`Vertex.connected`.

        Returns
        -------
        None : :class:`NoneType`

        """

        if getattr(cls, 'connected', False):
            return

        for v1, v2 in cls.connections:
            v1.connected.append(v2)
            v2.connected.append(v1)
        cls.connected = True



"""
Defines optimization functions which use MOPAC.

"""

import os
import subprocess as sp
import psutil
import time
import logging
import rdkit.Chem.AllChem as rdkit
from uuid import uuid4

logger = logging.getLogger(__name__)


def mopac_opt(mol, mopac_path, settings=None):
    """
    Optimizes the molecule using MOPAC.

    This function runs an optimization. It is possible to provide
    different options, which correspond to the input keywords from
    MOPAC:

    http://openmopac.net/Manual/index.html

    Parameters
    ----------
    mol : :class:`.Molecule`
        The molecule to be optimized.

    mopac_path : :class:`str`
        The full path to the MOPAC suite on the user's
        machine. For example, in a default MacOS installation the
        folder will probably be something like
        ``/opt/mopac/MOPAC2016.exe``.

    settings: :class:`dict`, optional
        A dictionary which maps the names of the optimization
        parameters to their values. Valid values are:

            'hamiltonian' : :class:`str` (default = ``'PM7'``)
                A series of different methods can be selected:
                PM7, PM6, AM1, CIS (CISD, CISDT), MNDO, RM1, etc..

                PM7 is the latest version of the reparametrization of
                NDDO theory, where all the atomic and diatomic
                parameters were re-optimized / updated from PM6 [#]_.

            'method' : :class:`str` (default = ``'OPT'``)
                The default calculation consists of a geometry
                optimization. You can run single point calculations
                (SCF) or transition search algorithms (TS). Refer to
                the MOPAC website for specific keywords.

            'gradient' : :class:`float` (default = ``0.01``)
                The gradient at which the geometry optimization reaches
                the convergence criteria (``kcal`` / ``mol`` /
                ``Angstrom``). For small system and high precision
                work, ``0.01`` is recommended.

            'eps' : :class:`float` (default = ``80.1``)
                Sets the dielectric constant for the solvent. Presence
                of this keyword will cause the COSMO (Conductor-like
                Screening Model) method to be used to approximate the
                effect of a solvent model surrounding the molecule.
                Solvents with a low dielectric constant are not likely
                to work well with this model. ``0`` means that the
                dielectric constant is not included in the calculation.
                ``80.1`` can be used to model a water environment at
                room temperature.

            'charge' : :class:`float` (default = ``0``)
                The charge of the system.

            'fileout' : :class:`str` (default = ``'PDBOUT'``)
                Determines the output file type.

            'timeout' : :class:`float` (default = ``172800``)
                The amount in seconds the optimization is allowed to
                run before being terminated. The default value is ``2``
                days or ``172,800`` seconds. ``None`` means there
                is no timeout.

    Returns
    -------
    None : :class:`NoneType`

    References
    ----------
    .. [#] http://openmopac.net/PM7_accuracy/PM7_accuracy.html

    """

    if settings is None:
        settings = {}

    vals = {
            'hamiltonian': 'PM7',
            'method': 'OPT',
            'gradient': 0.01,
            'eps': 80.1,
            'charge': 0,
            'fileout': 'PDBOUT',
            'timeout': 172800,
            }
    vals.update(settings)

    mol._file = '{}.mol'.format(uuid4().int)

    # First write a .mol file of the molecule.
    mol.write(mol._file)
    # MOPAC requires a ``.mop`` file as input. This creates a ``.mop``
    # file holding the molecule.
    _create_mop(mol, vals)
    # Run the optimization
    _run_mopac(mol, mopac_path, settings)
    # Update the rdkit mol info with the ``.pdb`` file generated from
    # the MOPAC run
    _convert_mopout_to_mol(mol)


def _run_mopac(mol, mopac_path, settings, timeout=7200):

    name, ext = os.path.splitext(mol._file)
    mop_file = name + '.mop'

    print("", time.ctime(time.time()),
          'Running MOPAC - {}.'.format(mol.name), sep='\n')

    # To run MOPAC a command is issued to the console via
    # ``subprocess.Popen``. The command is the full path of the
    # ``mopac`` program.
    file_root, ext = os.path.splitext(mop_file)
    opt_cmd = [mopac_path, file_root]
    opt_proc = psutil.Popen(opt_cmd, stdout=sp.PIPE,
                            stderr=sp.STDOUT,
                            universal_newlines=True)

    try:
        proc_out, _ = opt_proc.communicate(timeout=timeout)
    except sp.TimeoutExpired:
        logger.warning('\nMinimization took too long and was terminated '
                       'by force - {}\n'.format(mol.name))
        _kill_mopac(mol)

    return


def _kill_mopac(mol):
    """
    Kills an in-progress MOPAC run.

    To kill a MOPAC run, a file with the molecule's name and a ``.end``
    extension is written.

    Parameters
    ----------
    mol : :class:`.Molecule`
        The molecule being optimized.

    Returns
    -------
    None : :class:`NoneType`

    """
    name, ext = os.path.splitext(mol._file)
    end_file = name + '.end'

    with open(end_file, 'w') as end:
        end.write('SHUT')


def _mop_line(settings):
    """
    Formats `settings` into a MOPAC input string.

    Parameters
    ----------
    settings : :class:`dict`
        Dictionary defined in :func:`mopac_opt`, where all the run
        details are defined.

    Returns
    -------
    :class:`str`
        String containing all the MOPAC keywords correctly formatted
        for the input file.

    """

    # Generate an empty string
    mopac_run_str = ""

    # Add Hamiltonian info
    mopac_run_str = mopac_run_str + settings['hamiltonian']
    # Add method and GNORM if 'OPT' otherwise just print the method
    if settings['method'] == 'OPT':
        gnorm_info = ' GNORM={} '.format(settings['gradient'])
        mopac_run_str = mopac_run_str + " " + settings['method'] + gnorm_info
    else:
        mopac_run_str = mopac_run_str + settings['method']
    # Add EPS info
    eps_info = ' EPS={} '.format(settings['eps'])
    mopac_run_str = mopac_run_str + eps_info
    # Add Charge info
    charge_info = ' CHARGE={} '.format(settings['charge'])
    mopac_run_str = mopac_run_str + charge_info
    # Add fileout info
    mopac_run_str = mopac_run_str + " " + settings['fileout']

    # Add the let keyword avoiding the crash of MOPAC
    mopac_run_str = mopac_run_str + " LET "

    return mopac_run_str


def _create_mop(mol, settings):
    """
    Creates the ``.mop`` file holding the molecule to be optimized.

    Parameters
    ----------
    mol : :class:`.Molecule`
        The molecule which is to be optimized. Its molecular
        structure file is converted to a ``.mop`` file. The original
        file is also kept.

    settings : :class:`dict`
        Dictionary defined in :func:`mopac_opt`, where all the run
        details are defined.

    Returns
    -------
    :class:`str`
        The full path of the newly created ``.mop`` file.

    """
    name, ext = os.path.splitext(mol._file)
    mop_file = name + '.mop'
    mol = mol.mol

    logger.info('\nCreating .mop file - {}.'.format(mol.name))

    # Generate the mop file containing the MOPAC run info
    with open(mop_file, 'w') as mop:
        # line for the run info
        mop.write(_mop_line(settings) + "\n")
        # line with the name of the molecule
        mop.write(name + "\n\n")

        # print the structural info
        for atom in mol.GetAtoms():
            atom_id = atom.GetIdx()
            atom_symbol = atom.GetSymbol()
            x, y, z = mol.GetConformer().GetAtomPosition(atom_id)
            atom_info = f"{atom_symbol}   {x}   +1  {y}   +1  {z}   +1 \n"
            mop.write(atom_info)

    return mop_file


def _convert_mopout_to_mol(mol):
    """
    Updates the molecular structure if the optimization is successful.

    Takes the ``.pdb`` file of the neutral file generated from the
    MOPAC run and initializes a new ``rdkit`` molecule with those
    coordinates. `mol` is then updated to hold the new molecule.

    Parameters
    ----------
    mol : :class:`.Molecule`
        The macromolecule being optimized. The ``.pdb`` file holding
        its optimized structure is converted to a rdkit molecule.

    Returns
    -------
    None : :class:`NoneType`

    """
    name, ext = os.path.splitext(mol._file)
    pdb_file = name + ".pdb"

    logger.info("\nUpdating molecule with MOPAC optimized "
                "one - {}.\n".format(mol.name))

    new_mol = rdkit.MolFromPDBFile(pdb_file,
                                   sanitize=False,
                                   removeHs=False)
    # Updating the macro_mol.mol infos with the new mol
    mol.mol = new_mol



"""
Defines optimization functions which use MacroModel.

"""

import os
import subprocess as sp
import time
import rdkit.Chem.AllChem as rdkit
import warnings
import psutil
import re
from uuid import uuid4
import logging
import gzip

from ..utilities import MAEExtractor, flatten


logger = logging.getLogger(__name__)


class _ConversionError(Exception):
    def __init__(self, message):
        self.message = message


class _PathError(Exception):
    def __init__(self, message):
        self.message = message


class _ForceFieldError(Exception):
    def __init__(self, message):
        self.message = message


class _OptimizationError(Exception):
    def __init__(self, message):
        self.message = message


class _LewisStructureError(Exception):
    def __init__(self, message):
        self.message = message


def macromodel_opt(mol,
                   macromodel_path,
                   settings=None,
                   md=None,
                   conformer=-1):
    """
    Optimizes the molecule using MacroModel.

    This function runs a restricted optimization. The structures of the
    building blocks are frozen and only the new bonds formed between
    building blocks during assembly are optimized.

    Parameters
    ----------
    mol : `.Molecule`
        The molecule who's structure must be optimized.

    macromodel_path : :class:`str`
        The full path of the Schrodinger suite within the user's
        machine. For example, on a Linux machine this may be something
        like ``'/opt/schrodinger2017-2'``.

    settings : :class:`dict`, optional
        A dictionary which maps the names of optimization parameters to
        their values. Valid values are:

            'restricted' : :class:`bool` (default = ``True``)
                If ``False`` then all bonds are optimized, not just the
                ones created during macromolecular assembly. If
                ``True`` then a regular optimization is performed. If
                ``'both'`` then a restricted optimization is performed
                first, followed by a regular optimization.

            'timeout' : :class:`float` (default = ``None``)
                The amount in seconds the optimization is allowed to
                run before being terminated. ``None`` means there is no
                timeout.

            'force_field' : :class:`int` (default = ``16``)
                The number of the force field to be used.

            'max_iter' : :class:`int` (default = ``2500``)
                The maximum number of iterations done during the
                optimization.

            'gradient' : :class:`float` (default = ``0.05``)
                The gradient at which optimization is stopped.

            'md' : :class:`bool` (default = ``False``)
                Toggles whether a MD conformer search should be
                performed.

        Only values which need to be changed from the default need to
        be specified. For exmaple,

        .. code-block:: python

            macromodel_opt(mol, 'path', {'max_iter' : 10})


    md : :class:`dict`, optional
        A dictionary holding settings for the MD conformer search.
        This parameter is used in the same way as `settings` except
        the values effect the MD only. See docstring of
        :func:`_macromodel_md_opt` for valid values.

    conformer : :class:`int`, optional
        The id of the conformer to be optimized.

    Returns
    -------
    None : :class:`NoneType`

    """

    if settings is None:
        settings = {}
    if md is None:
        md = {}

    vals = {
             'restricted': True,
             'timeout': None,
             'force_field': 16,
             'max_iter': 2500,
             'gradient': 0.05,
             'md': False
            }
    vals.update(settings)

    try:
        mol._file = '{}.mol'.format(uuid4().int)
        # First write a .mol file of the molecule.
        mol.write(mol._file, conformer)
        # MacroModel requires a ``.mae`` file as input. This creates a
        # ``.mae`` file holding the molecule.
        _create_mae(mol, macromodel_path)
        # generate the ``.com`` file for the MacroModel run.
        _generate_com(mol, vals)
        # Run the optimization.
        _run_bmin(mol, macromodel_path, vals['timeout'])
        # Get the ``.maegz`` file output from the optimization and
        # convert it to a ``.mae`` file.
        _convert_maegz_to_mae(mol)
        mol.update_from_mae(mol._file.replace('.mol', '.mae'),
                            conformer)

        if vals['restricted'] == 'both':
            new_vals = dict(vals)
            new_vals['md'] = False
            new_vals['restricted'] = False
            macromodel_opt(mol=mol,
                           macromodel_path=macromodel_path,
                           settings=new_vals,
                           md={})

        if vals['md']:
            _macromodel_md_opt(mol,
                               macromodel_path,
                               md,
                               conformer)

    except _ForceFieldError as ex:
        # If OPLS_2005 has been tried already - record an exception.
        if vals['force_field'] == 14:
            raise ex

        # If OPLSE_2005 has not been tried - try it.
        logger.warning(('Minimization with OPLS3 failed on "{}". '
                        'Trying OPLS_2005.').format(mol.name))

        vals['force_field'] = 14
        return macromodel_opt(mol,
                              macromodel_path,
                              vals,
                              md,
                              conformer)


def macromodel_cage_opt(mol,
                        macromodel_path,
                        settings=None,
                        md=None,
                        conformer=-1):
    """
    Optimizes the molecule using MacroModel.

    This function runs a restricted optimization. The structures of the
    building blocks are frozen and only the new bonds formed between
    building blocks during assembly are optimized.

    This function differes from `macromodel_opt` in that it checks the
    number of windows the `macro_mol` has before running the MD. The MD
    is only run all windows are found  (i.e. the cage is not
    collapsed).

    Parameters
    ----------
    mol : `.Molecule`
        The molecule who's structure must be optimized.

    macromodel_path : :class:`str`
        The full path of the Schrodinger suite within the user's
        machine. For example, on a Linux machine this may be something
        like ``'/opt/schrodinger2017-2'``.

    settings : :class:`dict`, optional
        A dictionary which maps the names of optimization parameters to
        their values. Valid values are:

            'restricted' : :class:`bool` (default = ``True``)
                If ``False`` then all bonds are optimized, not just the
                ones created during macromolecular assembly. If
                ``True`` then a regular optimization is performed. If
                ``'both'`` then a restricted optimization is performed
                first, followed by a regular optimization.

            'timeout' : :class:`float` (default = ``None``)
                The amount in seconds the optimization is allowed to
                run before being terminated. ``None`` means there is no
                timeout.

            'force_field' : :class:`int` (default = ``16``)
                The number of the force field to be used.

            'max_iter' : :class:`int` (default = ``2500``)
                The maximum number of iterations done during the
                optimization.

            'gradient' : :class:`float` (default = ``0.05``)
                The gradient at which optimization is stopped.

            'md' : :class:`bool` (default = ``False``)
                Toggles whether a MD conformer search should be
                performed.

        Only values which need to be changed from the default need to
        be specified. For exmaple,

        .. code-block:: python

            macromodel_opt(mol, 'path', {'max_iter' : 10})


    md : :class:`dict`, optional
        A dictionary holding settings for the MD conformer search.
        This parameter is used in the same way as `settings` except
        the values effect the MD only. See docstring of
        :func:`_macromodel_md_opt` for valid values.

    conformer : :class:`int`, optional
        The id of the conformer to be optimized.

    Returns
    -------
    None : :class:`NoneType`

    """

    if settings is None:
        settings = {}
    if md is None:
        md = {}

    vals = {
             'restricted': True,
             'timeout': None,
             'force_field': 16,
             'max_iter': 2500,
             'gradient': 0.05,
             'md': False
            }
    vals.update(settings)

    try:
        mol._file = '{}.mol'.format(uuid4().int)
        # First write a .mol file of the molecule.
        mol.write(mol._file, conformer)
        # MacroModel requires a ``.mae`` file as input. This creates a
        # ``.mae`` file holding the molecule.
        _create_mae(mol, macromodel_path)
        # generate the ``.com`` file for the MacroModel run.
        _generate_com(mol, vals)
        # Run the optimization.
        _run_bmin(mol, macromodel_path, vals['timeout'])
        # Get the ``.maegz`` file output from the optimization and
        # convert it to a ``.mae`` file.
        _convert_maegz_to_mae(mol)
        mol.update_from_mae(mol._file.replace('.mol', '.mae'),
                            conformer)

        if vals['restricted'] == 'both':
            new_vals = dict(vals)
            new_vals['md'] = False
            new_vals['restricted'] = False
            macromodel_opt(mol=mol,
                           macromodel_path=macromodel_path,
                           settings=new_vals,
                           md={})

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            windows = mol.windows(conformer)
            if windows is not None:
                all_windows = (len(windows) ==
                               mol.topology.n_windows)

                if vals['md'] and all_windows:
                    _macromodel_md_opt(mol,
                                       macromodel_path,
                                       md,
                                       conformer)

    except _ForceFieldError as ex:
        # If OPLS_2005 has been tried already - record an exception.
        if vals['force_field'] == 14:
            raise ex

        # If OPLSE_2005 has not been tried - try it.
        logger.warning(('Minimization with OPLS3 failed on "{}". '
                        'Trying OPLS_2005.').format(mol.name))

        vals['force_field'] = 14
        return macromodel_cage_opt(mol,
                                   macromodel_path,
                                   vals,
                                   md,
                                   conformer)


def _macromodel_md_opt(mol,
                       macromodel_path,
                       settings=None,
                       conformer=-1):
    """
    Runs a MD conformer search on `mol`.

    Parameters
    ----------
    mol : `.Molecule`
        The molecule who's structure must be optimized.

    macromodel_path : :class:`str`
        The full path of the Schrodinger suite within the user's
        machine. For example, on a Linux machine this may be something
        like ``'/opt/schrodinger2017-2'``.

    settings : :class:`dict`, optional
        Each key is a string describing an MD parameter. The value is
        the corresponding value. Only values which need to be changed
        from the default need to be specified. The allowed parameters
        are:

            'timeout' : :class:`float` (default = ``None``)
                The amount in seconds the MD is allowed to run before
                being terminated. ``None`` means there is no timeout.

            'force_field' : :class:`int` (default = ``16``)
                The number of the force field to be used.

            'temp' : :class:`float` (default = ``300``)
                The temperature in Kelvin at which the MD is run.

            'confs' : :class:`int` (default = ``50``)
                The number of conformers sampled and optimized from the
                MD.

            'time_step' : :class:`float` (default = ``1.0``)
                The time step in ``fs`` for the MD.

            'eq_time' : :class:`float` (default = ``10``)
                The equilibriation time in ``ps`` before the MD is run.

            'sim_time' : :class:`float` (default = ``200``)
                The simulation time in ``ps`` of the MD.

            'max_iter' : :class:`int` (default = ``2500``)
                The maximum number of iterations done during the
                optimization.

            'gradient' : float (default = ``0.05``)
                The gradient at which optimization is stopped.

    conformer : :class:`int`, optional
        The id of the conformer to be optimized.

    Returns
    -------
    None : :class:`NoneType`

    """

    if settings is None:
        settings = {}

    vals = {
               'timeout': None,
               'force_field': 16,
               'temp': 300,
               'confs': 50,
               'time_step': 1.0,
               'eq_time': 10,
               'sim_time': 200,
               'max_iter': 2500,
               'gradient': 0.05
              }

    vals.update(settings)

    logger.info('Running MD on "{}".'.format(mol.name))
    try:
        mol._file = '{}.mol'.format(uuid4().int)
        # First write a .mol file of the molecule.
        mol.write(mol._file, conformer)
        # MacroModel requires a ``.mae`` file as input. This creates a
        # ``.mae`` file holding the molecule.
        _create_mae(mol, macromodel_path)
        # Generate the ``.com`` file for the MacroModel MD run.
        _generate_md_com(mol, vals)
        # Run the optimization.
        _run_bmin(mol, macromodel_path, vals['timeout'])
        # Extract the lowest energy conformer into its own .mae file.
        conformer_mae = MAEExtractor(mol._file).path
        mol.update_from_mae(conformer_mae, conformer)

    except _ForceFieldError as ex:
        # If OPLS_2005 has been tried already - record an exception.
        if vals['force_field'] == 14:
            raise ex
        # If OPLSE_2005 has not been tried - try it.
        logger.warning(('Minimization with OPLS3 failed on "{}". '
                        'Trying OPLS_2005.').format(mol.name))

        vals['force_field'] = 14
        return _macromodel_md_opt(mol,
                                  macromodel_path,
                                  vals,
                                  conformer)


def _run_bmin(macro_mol, macromodel_path, timeout):

    logger.info('Running bmin on "{}".'.format(macro_mol.name))

    # To run MacroModel a command is issued to the console via
    # ``subprocess.Popen``. The command is the full path of the
    # ``bmin`` program. ``bmin`` is located in the Schrodinger
    # installation folder.
    file_root, ext = os.path.splitext(macro_mol._file)
    log_file = file_root + '.log'
    opt_app = os.path.join(macromodel_path, "bmin")
    # The first member of the list is the command, the following ones
    # are any additional arguments.

    opt_cmd = [opt_app, file_root, "-WAIT", "-LOCAL"]

    incomplete = True
    while incomplete:
        opt_proc = psutil.Popen(opt_cmd,
                                stdout=sp.PIPE,
                                stderr=sp.STDOUT,
                                universal_newlines=True)
        try:
            proc_out, _ = opt_proc.communicate(timeout=timeout)

        except sp.TimeoutExpired:
            logger.warning(('Minimization took too long'
                            ' and was terminated '
                            f'by force on "{macro_mol.name}".'))
            _kill_bmin(macro_mol, macromodel_path)
            proc_out = ""

        logger.debug(
            f'Output of bmin on "{macro_mol.name}" was: {proc_out}.')

        with open(log_file, 'r') as log:
            log_content = log.read()

        # Check the log for error reports.
        if ("termination due to error condition           21-" in
           log_content):
            raise _OptimizationError(("`bmin` crashed due to"
                                      " an error condition. "
                                      "See .log file."))

        if ("FATAL do_nosort_typing: NO MATCH found for atom " in
           log_content):
            raise _ForceFieldError(
                            'The log implies the force field failed.')

        if (("FATAL gen_lewis_structure(): could not find best Lewis"
             " structure") in log_content and
            ("skipping input structure  due to "
             "forcefield interaction errors") in log_content):
            raise _LewisStructureError(
                    '`bmin` failed due to poor Lewis structure.')

        # If optimization fails because a wrong Schrodinger path was
        # given, raise.
        if 'The system cannot find the path specified' in proc_out:
            raise _PathError(('Wrong Schrodinger path supplied to'
                              ' `macromodel_opt` function.'))

        # If optimization fails because the license is not found, rerun
        # the function.
        if _license_found(proc_out, macro_mol):
            incomplete = False

    # Make sure the .maegz file created by the optimization is present.
    maegz = file_root + '-out.maegz'
    _wait_for_file(maegz)
    if not os.path.exists(log_file) or not os.path.exists(maegz):
        raise _OptimizationError(('The .log and/or .maegz '
                                  'files were not created by '
                                  'the optimization.'))


def _kill_bmin(macro_mol, macromodel_path):
    name, ext = os.path.splitext(macro_mol._file)
    name = re.split(r'\\|/', name)[-1]
    app = os.path.join(macromodel_path, 'jobcontrol')
    cmd = [app, '-stop', name]

    incomplete = True
    while incomplete:
        out = sp.run(cmd, stdout=sp.PIPE,
                     stderr=sp.STDOUT, universal_newlines=True)

        # If no license if found, keep re-running the function until it
        # is.
        if _license_found(out.stdout):
            incomplete = False

    # This loop causes the function to wait until the job has been
    # killed via job control. This means the output files will have
    # been written by the time the function exits. Essentially the loop
    # continues until the job is no longer found by
    # "./jobcontrol -list"
    cmd = [app, '-list']
    output = name
    start = time.time()
    while name in output:
        output = sp.run(cmd, stdout=sp.PIPE, stderr=sp.STDOUT,
                        universal_newlines=True).stdout
        if time.time() - start > 600:
            break


def _license_found(output, mol=None):
    """
    Checks to see if minimization failed due to a missing license.

    The user can be notified of this in one of two ways. Sometimes the
    output of the submission contains the message informing that the
    license was not found and in other cases it will be the log file.
    This function checks both of these sources for this message.

    Parameters
    ----------
    output : :class:`str`
        The output from submitting the minimization of the structure
        to the ``bmin`` program.

    mol : :class:`.Molecule`, optional
        The molecule being optimized. If the ``.log`` file is not to
        be checked, the default ``None`` should be used.

    Returns
    -------
    :class:`bool`
        ``True`` if the license was found. ``False`` if the
        minimization did not occur due to a missing license.

    """

    if 'Could not check out a license for mmlibs' in output:
        return False
    if mol is None:
        return True

    # To check if the log file mentions a missing license file open the
    # the log file and scan for the apporpriate string.

    # Check if the file exists first. If not, this is often means the
    # calculation must be redone so return False anyway.
    log_file_path = mol._file.replace('mol', 'log')
    with open(log_file_path, 'r') as log_file:
        log_file_content = log_file.read()

    if 'Could not check out a license for mmlibs' in log_file_content:
        return False

    return True


def _com_line(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9):
    return (" {:<5}{:>7}{:>7}{:>7}{:>7}{:>11.4f}{:>11.4f}"
            "{:>11.4f}{:>11.4f}").format(arg1, arg2, arg3, arg4,
                                         arg5, arg6, arg7, arg8, arg9)


def _generate_com(mol, settings):
    """
    Create a ``.com`` file for a MacroModel optimization.

    The created ``.com`` file fixes all bond parameters which were not
    added during assembly. This means all bond distances, bond angles
    and torsional angles are fixed, except for cases where it involves
    a bond added during assembly of the macromolecule.

    This fixing is implemented by creating a ``.com`` file with various
    "FX" commands written within its body.

    Parameters
    ----------
    mol : :class:`.Molecule`
        The molecule which is to be optimized.

    settings : :class:`dict`
        A dictionary of settings for the optimization. See
        :func:`macromodel_opt` documentation.

    Returns
    -------
    None : :class:`NoneType`

    """

    logger.debug('Creating .com file for "{}".'.format(mol.name))

    # This is the body of the ``.com`` file. The line that begins and
    # ends with exclamation lines is replaced with the various commands
    # that fix bond distances and angles.
    main_string = "\n".join([
        _com_line('MMOD', 0, 1, 0, 0, 0, 0, 0, 0),

        _com_line('FFLD',
                  settings['force_field'], 1, 0, 0, 1, 0, 0, 0),

        _com_line('BGIN', 0, 0, 0, 0, 0, 0, 0, 0),
        _com_line('READ', 0, 0, 0, 0, 0, 0, 0, 0),
        "!!!BLOCK_OF_FIXED_PARAMETERS_COMES_HERE!!!",
        _com_line('CONV', 2, 0, 0, 0, settings['gradient'], 0, 0, 0),
        _com_line('MINI', 1, 0, settings['max_iter'], 0, 0, 0, 0, 0),
        _com_line('END', 0, 1, 0, 0, 0, 0, 0, 0)])

    # Create a path for the ``.com`` file. It is the same as that of
    # the structure file but with a ``.com`` extension. Get the path of
    # the ``.mae`` file and the output file in the same way.
    name, ext = os.path.splitext(mol._file)
    com_file = name + '.com'
    mae = name + '.mae'
    output = name + '-out.maegz'

    # This function adds all the lines which fix bond distances and
    # angles into ``main_string``.
    main_string = _fix_params_in_com_file(mol, main_string,
                                          settings['restricted'])

    # Writes the ``.com`` file.
    with open(com_file, "w") as com:
        # The first line hold the ``.mae`` file containing the molecule
        # to be optimized.
        com.write(str(mae + "\n"))
        # The second line holds the name of the output file of the
        # optimization.
        com.write(str(output + "\n"))
        # Next is the body of the ``.com`` file, held in
        # ``main_string``.
        com.write(main_string)


def _generate_md_com(macro_mol, settings):
    logger.debug('Creating .com file for "{}".'.format(macro_mol.name))

    main_string = "\n".join([
        _com_line('MMOD', 0, 1, 0, 0, 0, 0, 0, 0),

        _com_line('FFLD',
                  settings['force_field'], 1, 0, 0, 1, 0, 0, 0),

        _com_line('READ', 0, 0, 0, 0, 0, 0, 0, 0),
        _com_line('MDIT', 0, 0, 0, 0, settings['temp'], 0, 0, 0),

        _com_line('MDYN', 0, 0, 0, 0, settings['time_step'],
                  settings['eq_time'], settings['temp'], 0),

        _com_line('MDSA', settings['confs'], 0, 0, 0, 0, 0, 1, 0),

        _com_line('MDYN', 1, 0, 0, 0, settings['time_step'],
                  settings['sim_time'], settings['temp'], 0),

        _com_line('WRIT', 0, 0, 0, 0, 0, 0, 0, 0),
        _com_line('RWND', 0, 1, 0, 0, 0, 0, 0, 0),
        _com_line('BGIN', 0, 0, 0, 0, 0, 0, 0, 0),
        _com_line('READ', -2, 0, 0, 0, 0, 0, 0, 0),
        _com_line('CONV', 2, 0, 0, 0, settings['gradient'], 0, 0, 0),
        _com_line('MINI', 1, 0, settings['max_iter'], 0, 0, 0, 0, 0),
        _com_line('END', 0, 1, 0, 0, 0, 0, 0, 0)])

    name, ext = os.path.splitext(macro_mol._file)
    com_file = name + '.com'
    mae = name + '.mae'
    output = name + '-out.maegz'

    # Generate the com file containing the info for the run
    with open(com_file, "w") as com:
        # name of the macromodel file
        com.write(str(mae + "\n"))
        # name of the output file
        com.write(str(output + "\n"))
        # details of the macromodel run
        com.write(main_string)


def _create_mae(mol, macromodel_path):
    """
    Creates the ``.mae`` file holding the molecule to be optimized.

    Parameters
    ----------
    mol : :class:`.Molecule`
        The molecule which is to be optimized. Its molecular
        structure file is converted to a ``.mae`` file. The original
        file is also kept.

    macromodel_path : :class:`str`
        The full path of the Schrodinger suite within the user's
        machine. For example, on a Linux machine this may be something
        like ``'/opt/schrodinger2017-2'``.

    Returns
    -------
    :class:`str`
        The full path of the newly created ``.mae`` file.

    """

    _, ext = os.path.splitext(mol._file)

    logger.debug(f'Converting {ext} of "{mol.name}" to .mae.')

    # Create the name of the new ``.mae`` file. It is the same as the
    # original structure file, including the same path. Only the
    # extensions are different.
    mae_file = mol._file.replace(ext, '.mae')
    _structconvert(mol._file, mae_file, macromodel_path)
    return mae_file


def _convert_maegz_to_mae(mol):
    """
    Converts a ``.maegz`` file to a ``.mae`` file.

    Parameters
    ----------
    mol : :class:`.Molecule`
        The molecule being optimized. The ``.maegz`` file holding
        its optimized structure is converted to a ``.mae`` file. Both
        versions are kept.

    Returns
    -------
    None : :class:`NoneType`

    """

    logger.debug(f'Converting .maegz of "{mol.name}" to .mae.')
    name, ext = os.path.splitext(mol._file)
    # ``out`` is the full path of the optimized ``.mae`` file.
    maegz = name + '-out.maegz'
    # Replace extensions to get the names of the various files.
    mae = name + '.mae'

    gz_file = gzip.open(maegz)
    with open(mae, 'wb') as f:
        f.write(gz_file.read())
    gz_file.close()


def _structconvert(iname, oname, macromodel_path):

    convrt_app = os.path.join(macromodel_path,
                              'utilities',
                              'structconvert')
    convrt_cmd = [convrt_app, iname, oname]

    incomplete = True
    while incomplete:

        # Execute the file conversion.
        try:
            convrt_return = sp.run(convrt_cmd,
                                   stdout=sp.PIPE,
                                   stderr=sp.STDOUT,
                                   universal_newlines=True)

        # If conversion fails because a wrong Schrodinger path was
        # given, raise.
        except FileNotFoundError:
            raise _PathError(('Wrong Schrodinger path supplied to'
                              ' `structconvert` function.'))

        if 'File does not exist' in convrt_return.stdout:
            raise _ConversionError(
                    (f'structconvert input file, {iname}, missing. '
                     f'Console output was {convrt_return.stdout}'))

        # If no license if found, keep re-running the function until it
        # is.
        if _license_found(convrt_return.stdout):
            incomplete = False

    # If force field failed, raise.
    if 'number 1' in convrt_return.stdout:
        raise _ForceFieldError(convrt_return.stdout)

    _wait_for_file(oname)
    if not os.path.exists(oname):
        raise _ConversionError(
         (f'Conversion output file {oname} was not found.'
          f' Console output was {convrt_return.stdout}.'))

    return convrt_return


def _fix_params_in_com_file(mol, main_string, restricted):
    """
    Adds lines to the ``.com`` body fixing bond distances and angles.

    For each bond distance, bond angle and torisional angle that does
    not involve a bond created during assembly a "FX" command is
    added to the string holding holding the body of the ``.com`` file.

    These lines replace the filler line in the main string.

    Parameters
    ----------
    mol : :class:`.Molecule`
        The molecule which is to be optimized.

    main_string : :class:`str`
        The body of the ``.com`` file which is to have fix commands
        added.

    restricted : :class:`bool`
        When ``False`` the block containing instructions to fix
        molecular parameters is not added to the ``.com`` file.

    Returns
    -------
    :class:`str`
        A string holding the body of the ``.com`` file with
        instructions to fix the various bond distances and angles as
        described in the docstring.

    """

    # Make a string to hold all of the ``FX`` lines.
    fix_block = ""

    # If `restricted` is ``False`` do not add a fix block.
    if not restricted:
        return main_string.replace(("!!!BLOCK_OF_FIXED_PARAMETERS_"
                                    "COMES_HERE!!!\n"), fix_block)
    # Add lines that fix the bond distance.
    fix_block = _fix_distance_in_com_file(mol, fix_block)
    # Add lines that fix the bond angles.
    fix_block = _fix_bond_angle_in_com_file(mol, fix_block)
    # Add lines that fix the torsional angles.
    fix_block = _fix_torsional_angle_in_com_file(mol, fix_block)

    return main_string.replace(("!!!BLOCK_OF_FIXED_PARAMETERS_"
                                "COMES_HERE!!!\n"), fix_block)


def _fix_distance_in_com_file(mol, fix_block):
    """
    Adds lines fixing bond distances to ``.com`` body string.

    Only bond distances which do not involve bonds created during
    assembly are fixed.

    Parameters
    ----------
    mol : :class:`.Molecule`
        The molecule to be optimized.

    fix_block : :class:`str`
        The string holding all the lines containing fix commands for
        the ``.com`` file.

    Returns
    -------
    :class:`str`
        A string holding lines containg fix commands for the ``.com``
        file. The string has the lines fixing bond distances added to
        it by this function.

    """

    bonder_ids = set(flatten(mol.bonder_ids))

    # Go through all the bonds in the rdkit molecule. If the bond
    # is not between bonder atoms add a fix line to the ``fix_block``.
    # If the bond does invovle two bonder atoms go to the next bond.
    # This is because a bond between 2 bonder atoms was added during
    # assembly and should therefore not be fixed.
    for bond in mol.mol.GetBonds():
        atom1 = bond.GetBeginAtom()
        atom2 = bond.GetEndAtom()

        if (atom1.GetIdx() in bonder_ids and
           atom2.GetIdx() in bonder_ids):
            continue

        # Make sure that the indices are increased by 1 in the .mae
        # file.
        atom1_id = atom1.GetIdx() + 1
        atom2_id = atom2.GetIdx() + 1
        fix_block += (_com_line('FXDI', atom1_id, atom2_id, 0, 0,
                                99999, 0, 0, 0) + '\n')

    return fix_block


def _fix_bond_angle_in_com_file(mol, fix_block):
    """
    Adds lines fixing bond angles to ``.com`` body string.

    All bond angles of the molecule are fixed.

    Parameters
    ----------
    mol : :class:`.Molecule`
        The molecule to be optimized.

    fix_block : :class:`str`
        The string holding all the lines containing fix commands for
        the ``.com`` file.

    Returns
    -------
    :class:`str`
        A string holding lines containg fix commands for the ``.com``
        file. The string has the lines fixing bond angles added to it
        by this function.

    """

    # Create a substructure consisting of 3 dummy atoms bonded with 3
    # dummy bonds. This substructure will match with any 3 atoms which
    # are bonded together with any combination of bonds. These 3 atoms
    # will therefore have a bond angle.
    ba_mol = rdkit.MolFromSmarts('[*]~[*]~[*]')

    # Get the indices of all atoms which have a bond angle.
    # ``ba_atoms`` is a tuple of tuples of the form ((1,2,3), (4,5,6),
    # (7,8,9), ...). Each inner tuple holds the indicies of the atoms
    # which form a bond angle.
    ba_atoms = mol.mol.GetSubstructMatches(ba_mol)

    for atom_ids in ba_atoms:
        atom_ids = [i+1 for i in atom_ids]
        fix_block += (_com_line('FXBA', *atom_ids,
                                99999, 0, 0, 0, 0) + '\n')

    return fix_block


def _fix_torsional_angle_in_com_file(mol, fix_block):
    """
    Adds lines fixing torsional bond angles to ``.com`` body string.

    All torsional angles of the molecule are fixed.

    Parameters
    ----------
    macro_mol : :class:`.Molecule`
        The molecule to be optimized.

    fix_block : :class:`str`
        The string holding all the lines containing fix commands for
        the ``.com`` file.

    Returns
    -------
    :class:`str`
        A string holding lines containg fix commands for the ``.com``
        file. The string has the lines fixing bond angles added to it
        by this function.

    """

    # Create a substructure consisting of 4 dummy atoms bonded with 4
    # dummy bonds. This substructure will match with any 4 atoms which
    # are bonded together with any combination of bonds. These 4 atoms
    # will therefore have a torsinal angle.
    ta_mol = rdkit.MolFromSmarts('[*]~[*]~[*]~[*]')

    # Get the indices of all atoms which have a torsional angle.
    # ``ta_atoms`` as a tuple of tuples of the form ((1,2,3,4),
    # (4,5,6,7), ...). Each inner tuple holds the indicies of the atoms
    # which have a torsional angle.
    ta_atoms = mol.mol.GetSubstructMatches(ta_mol)

    # Apply the fix.
    for atom_ids in ta_atoms:
        atom_ids = [i+1 for i in atom_ids]
        fix_block += (_com_line('FXTA', *atom_ids,
                                99999, 361, 0, 0) + '\n')

    return fix_block


def _wait_for_file(file_name, timeout=10):
    """
    Stalls until a given file exists or `timeout` expires.

    Parameters
    ----------
    file_name : :class:`str`
        The full path of the file which should be waited for.

    timeout : :class:`int` or :class:`float`, optional
        The number of seconds before the function stops waiting and
        returns.

    Returns
    --------
    None : :class:`NoneType`

    """

    t_start = time.time()
    tick = 0
    while True:
        time_taken = time.time() - t_start
        if divmod(time_taken, 5)[0] == tick + 1:
            logger.warning('Waiting for "{}".'.format(file_name))
            tick += 1

        if os.path.exists(file_name) or time_taken > timeout:
            break



"""
Defines energy calculations via :class:`Energy`.

.. _`adding energy functions`:

Extending stk: Adding more energy functions.
--------------------------------------------

It may well be the case that more ways to calculate the energy of
molcules will need to be added. For example, if support for using some
external software to perform the calculations needs to be added.

In order to add a new function which calculates the energy, just add it
as a method to :class:`Energy`. There really aren't any requirements
beyond this, just make sure the energy value is what the method
returns.

If the energy method does not fit neatly into a single method, feel
free to split it up. Make sure that any helper methods are
private, i.e. that their names start with a leading underscore. Only
the main method which the user will use should be public.

To calculate the energy of some :class:`.Molecule` object the only
thing that is needed is calling one of the :class:`Energy` methods:

    >>>  molecule.energy.rdkit('uff')
    OUT: 16.501

Here, :meth:`Energy.rdkit` was used as an example and ``molecule``
is just some :class:`.Molecule` instance. Each :class:`.Molecule`
object has an :class:`Energy` instance in its :attr:`~.Molecule.energy`
attribute.

Each :class:`Energy` instance also has a :attr:`~Energy.values`
attribute, which stores the result of previous calculations. Using the
molecule from the previous example:

    >>>  molecule.energy.values
    OUT: {FunctionData('rdkit', forcefield='uff'): 16.501}

I.E. this indicates that :meth:`Energy.rdkit` was called with the
`forcefield` argument set to ``'uff'`` and the result was ``16.501``.

Calling any method of :class:`Energy` updates the dictionary
automatically. When adding a new method to the class, no mechanism for
updating the dictionary needs to be provided. Writing the method within
the class is enough for it to update the :attr:`~Energy.values`
dictionary when called.

Sometimes a function will require a parameter which does not affect the
outcome of the energy calculation. For example, to calculate the energy
using the MacroModel program, :meth:`Energy.macromodel` can be used:

.. code-block:: python

    molecule.energy.macromodel(forcefield=16,
                               macromodel_path='path/to/macromodel/dir')

This function requires the number of a forcefield (``16``) and the
directory where MacroModel is installed on the users computer
(``'path/to/macromodel/dir'``). However, the directory does not affect
the value of the calculated energy. When running:

    >>> molecule.energy.values

We want the output to be:

.. code-block:: python

    OUT: {FunctionData('rdkit', forcefield=uff): 16.501,
          FunctionData('macromodel', forcefield=16): 200}

(Assuming we are still dealing with the same ``molecule`` instance from
the :meth:`~Energy.rdkit` example, both calculated energies will be
kept in the dictionary.)

However, if we just define :meth:`Energy.macromodel` within
:class:`Energy`, and then run it:

.. code-block:: python

    molecule.energy.macromodel(forcefield=16,
                               macromodel_path='path/to/macromodel/dir')

The output of

    >>> molecule.energy.values

will be

.. code-block:: python

    OUT: {FunctionData('rdkit', forcefield=uff) : 16.501,
          FunctionData('macromolecule',
                       forcefield=16,
                       macromodel_path='path/to/macromodel/dir'): 200}

In order to make sure that the `macromodel_path` is excluded from the
key, decorate :meth:`Energy.macromodel` with the :func:`exclude()`
decorator. For example:

.. code-block:: python

    @exclude('macromodel_path')
    def macromodel(self, forcefield, macromodel_path):
        ...

Now the parameter `macromodel_path` will not form part of the key in
the :attr:`~Energy.values` dictionary. If there were 2 parameters you
wanted to exlude:

.. code-block:: python

    @exclude('exclude1', 'exclude2')
    def energy_func(include1, include2, exclude1, exclude2):
        ...

and so on.

Sometimes even this isn't enough to get the key to look exactly the way
we want. For exmple:

.. code-block:: python

    @exclude('force_e_calc')
    def formation(self,
                  func,
                  products,
                  building_blocks=None,
                  force_e_calc=False):
        ...

The `func` parameter of this function is a :class:`.FunctionData`
instance, which holds the data of one of the other :class:`Energy`
methods. This includes all data the method requires to run, even
software directories if needed. As a result, the key when running this
function may look like this:

.. code-block:: python

    FunctionData('pseudoformation',
                 building_blocks=None,
                 func=FunctionData('macromodel',
                                   forcefield=16,
                                   macromodel_path='/opt/schrodinger2017-2'))

Notice that the path of the MacroModel installation was kept nested in
the key of :meth:`Energy.formation`. This is undesirable. To make this
work properly, a completely custom mechanism for making the
key of the :meth:`Energy.formation` is necessary. To do this, define a
function in this module. For example:

.. code-block:: python

    def formation_key(fargs, fkwargs):
        ...

And set the :attr:`key` attribute of the energy method to the newly
defined function:

.. code-block:: python

    Energy.formation.key = formation_key

The `fargs` and `fkwargs` arguments are the arguments and keyword
arguments with which :meth:`Energy.formation` was called, including
`self`. The :func:`formation_key` function should return a
:class:`.FunctionData` instance which will act as the key. In our case
the function was defined so that the key is:

.. code-block:: python

    FunctionData('formation',
                 products=[],
                 building_blocks=None,
                 func=FunctionData('macromodel', forcefield=16))

Notes.
------

#. Make sure to use the `values` dictionary instead of running the same
   calculation repeatedly.

#. The automatic updating of the dictionary is  achieved by the
   :class:`EMeta` metaclass, :func:`EMethod` descriptor and
   :func:`e_logger` decorator. You do not need to worry about these,
   but information about how they work is provided in the docstring of
   :class:`EMeta`.

"""

import os
import rdkit.Chem.AllChem as rdkit
import subprocess as sp
import psutil
import copy
from uuid import uuid4
from types import MethodType
from functools import wraps
from inspect import signature as sig
import logging

from ..utilities import FunctionData
from ..optimization.mopac import mopac_opt


logger = logging.getLogger(__name__)


class EnergyError(Exception):
    """
    A class for errors in :class:`Energy` methods.

    """

    ...


class EMethod:
    """
    A descriptor for methods of :class:`Energy`.

    Attributes
    ----------
    func : :class:`function`
        The method which the descriptor acts as a getter for.

    """

    def __init__(self, func):
        """
        Initializes a :class:`EMethod` instance.

        Parameters
        ----------
        func : :class:`function`
            The method which the descriptor acts as a getter for.

        """

        self.func = func

    def __get__(self, obj, cls):
        """
        Returns a modified :attr:`func`.

        Attributes
        ----------
        obj : :class:`object`
            The object to which the method in :attr:`func` should be
            bound.

        cls : :class:`object`
            The class of `obj`.

        Returns
        -------
        :class:`BoundMethod`
            A decorated version of the method held in `self.func`. The
            difference is that when calling the method  now, it will
            automatically update the `values` attribute of `obj`.

        :class:`function`
            If the method in :attr:`func` is called as a class
            attribute rather than an instance attribute, return it
            instead of the descriptor.

        """
        # If the Energy method is accessed as a class attribute return
        # the unbound method.
        if obj is None:
            return self.func

        # When trying to access the Energy method as an instance
        # attribute returned a modified version of the method. The
        # method is modified so that after the method returns a value,
        # it is stored in the `values` dictionary of the Energy
        # instance.
        return e_logger(self.func, obj)


class EMeta(type):
    """
    A metaclass for :class:`Energy`.

    In conjuction with the :class:`EMethod` descriptor and the
    :func:`e_logger` decorator this class allows methods to
    automatically update :attr:`Energy.values` without explicitly being
    told to do so.

    Basically, this metaclass turns all methods of :class:`Energy` into
    descriptors of the :class:`EMethod` class. These descriptors return
    a decorated version of the original method defined in
    :class:`Energy`. The method is decorated with :func:`e_logger`
    decorator. Calling this decorated method makes it automatically
    update :attr:`Energy.values`.

    """

    def __new__(cls, cls_name, bases, cls_dict):
        """
        Turns all the public methods in `cls` into descriptors.

        """

        # Find all the methods defined in the class `cls` and replace
        # them with an ``Emethod`` descriptor.
        for name, func in cls_dict.items():

            # Don't replace any private methods.
            if name.startswith('_') or not callable(func):
                continue

            # Replace method with descriptor.
            cls_dict[name] = EMethod(func)

        return type.__new__(cls, cls_name, bases, cls_dict)


def e_logger(func, obj):
    """
    Turns `func` into a version which updates :attr`Energy.values`.

    Parameters
    ----------
    func : :class:`function`
        An :class:`Energy` method.

    obj : :class:`Energy`
        The :class:`Energy` object on which the method `func` was
        called.

    Returns
    -------
    :class:`types.MethodType`
        The function `func` bound to `obj` and modified so that when
        called the results update :attr:`Energy.values`.

    """

    @wraps(func)
    def inner(self, *args, **kwargs):

        # First get the result of the energy calculation.
        result = func(self, *args, **kwargs)

        # Next create FunctionData object to store the values of the
        # parameters used to run that calculation.
        key = func_key(func, (self,)+args, kwargs)

        # Update the `values` dictionary with the results of the
        # calculation.
        obj.values.update({key: result})
        # Return the result.
        return result

    # Make sure that the decorated function behaves like a typical
    # method and return.
    return MethodType(inner, obj)


def func_key(func, fargs=None, fkwargs=None):
    """
    Returns the key used in :attr:`Energy.values` for `func`.

    Parameters
    ----------
    func : :class:`function`
        The function whose results are to be stored in
        :attr:`Energy.values`.

    fargs : :class:`tuple`, optional
        The arguments passed to `func`.

    fkwargs : :class:`dict`, optional
        The keyword arguments passed to `func`.

    Returns
    -------
    :class:`.FunctionData`
        The :class:`.FunctionData` object representing the key when
        `func` is called with the arguments `fargs` and keyword
        arguments `fkwargs`.

    """

    if fargs is None:
        fargs = []
    if fkwargs is None:
        fkwargs = []

    # Check if the function has a `key` attribute. If it does use this
    # to get its key rather than the general purpose code written here.
    if hasattr(getattr(Energy, func.__name__, False), 'key'):
        return getattr(Energy, func.__name__).key(fargs, fkwargs)

    fsig = sig(func)
    # Get a dictionary of all the supplied parameters.
    bound = dict(fsig.bind_partial(*fargs, **fkwargs).arguments)
    # Get a dictionary of all the default initialized parameters.
    default = {key: value.default for key, value in
               dict(fsig.parameters).items() if
               key not in bound.keys()}
    # Combine the two sets of parameters and get rid of the `self`
    # parameter, if present.
    bound.update(default)
    if 'self' in bound.keys():
        bound.pop('self')
    # Remove any parameters that should not form key, listed in the
    # `exclude` attribute.
    if hasattr(func, 'exclude'):
        for key in func.exclude:
            bound.pop(key)

    # Return an FunctionData object representing the function and
    # chosen parameters.
    return FunctionData(func.__name__, **bound)


def exclude(*args):
    """
    A decorator to add the :attr`exclude` attribute to methods.

    Parameters
    ----------
    args :  :class:`tuple` of :class:`str`
        Holds the names parameters which are not to be used as part of
        the key for the :class:`Energy` method.

    Returns
    -------
    :class:`function`
        The function which has had the :attr:`exclude` attribute added.
        This is a :class:`list` holding the names of parameters, which
        are not part of the key.

    """

    def inner(func):
        func.exclude = args
        return func

    return inner


class Energy(metaclass=EMeta):
    """
    Handles all things related to a molecule's energy.

    An instance of this class will be placed in
    :attr:`.Molecule.energy` of each :class:`.Molecule` instance.

    Attributes
    ----------
    molecule : :class:`.Molecule`
        The molecule on which all the energy calculation are performed.

    values : :class:`dict`
        A :class:`dict` of the form

        .. code-block:: python

            values = {FunctionData('method1', param='a'): 34}

        Where the :class:`.FunctionData` instance identifies an
        :class:`Energy` method which was run and the parameters which
        were used, while the key holds the result of the calculation.

    """

    def __init__(self, molecule):
        """
        Initializes a :class:`Energy` instance.

        Parameters
        ----------
        molecules : :class:`.Molecule`
            The molecule on which all the energy calculation are
            performed.

        """

        self.molecule = molecule
        self.values = {}

    @exclude('force_e_calc')
    def formation(self,
                  func,
                  products,
                  building_blocks=None,
                  force_e_calc=False,
                  conformer=-1):
        """
        Calculates the formation energy.

        The formation energy is calculated under the assumption that
        the molecule in :attr:`molecule` is composed of the molecules
        in `building_blocks` and that during formation molecules in
        `products` are formed in addition to :attr:`molecule`.

        Parameters
        ----------
        func : :class:`.FunctionData`
            A :class:`.FunctionData` object which describes the method
            used to calculate the energies. For example:

            .. code-block:: python

                func  = FunctionData('rdkit', forcefield='uff')

        products : :class:`list`
            A :class:`list` of the form

            .. code-block:: python

                products = [(4, mol1), (2, mol2)]

            Where ``mol1`` and ``mol2`` are :class:`.Molecule` objects
            of molecules produced in addition to :attr:`molecule`
            during its formation reaction. The numbers represent the
            number of each molecule made per :attr:`molecule`.

        building_blocks : :class:`list`, optional
            A :class:`list` of the form

            .. code-block:: python

                building_blocks = [(2, mol1), (3, mol2)]

            Where ``mol1`` and ``mol2`` are :class:`.Molecule` objects
            which react to for :attr:`molecule`. The numbers represetnt
            the amount of each building block required to form 1
            :attr:`molecule`.

            This argument can be omitted when the formation energy of a
            :class:`.MacroMolecule` instance is being found, as they
            keep this data stored elsewhere already.

        force_e_calc : :class:`bool`, optional
            If this is ``True`` then all molecules in
            `building_blocks`, `products` and :attr:`molecule` will
            have their energies recalculated. Even if the energy values
            have already been found with the chosen forcefield and
            method. If ``False`` the energy is only calculated if the
            value has not already been found.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        :class:`float`
            The formation energy.

        """

        # Get the function used to calculate the energies.
        efunc = getattr(self, func.name)
        # Get the key of the function used to calculate the energies.
        fkey = func_key(efunc, None, func.params)

        # Recalculate energies if requested.
        if force_e_calc:
            for _, mol in products:
                getattr(mol.energy, func.name)(**func.params)

        # Get the total energy of the products.
        e_products = 0
        for n, mol in products:
            # If the energy has not been calculated already, calculate
            # it now.
            if fkey not in mol.energy.values.keys():
                getattr(mol.energy, func.name)(**func.params)

            e_products += n * mol.energy.values[fkey]

        eng = self.pseudoformation(func,
                                   building_blocks,
                                   force_e_calc,
                                   conformer)
        eng -= e_products
        return eng

    def pseudoformation(self,
                        func,
                        building_blocks=None,
                        force_e_calc=False,
                        conformer=-1):
        """
        Calculates the formation energy, sans other products.

        This is the formation energy if the energy of the other
        products of the reaction is not taken into account.

        Parameters
        ----------
        func : :class:`.FunctionData`
            A :class:`.FunctionData` object which describes the method
            used to calculate the energies. For example:

            .. code-block:: python

                func  = FunctionData('rdkit', forcefield='uff')

        building_blocks : :class:`list`, optional
            A :class:`list` of the form

            .. code-block:: python

                building_blocks = [(2, mol1), (3, mol2)]

            Where ``mol1`` and ``mol2`` are :class:`.Molecule` objects
            which react to for :attr:`molecule`. The numbers represetnt
            the amount of each building block required to form 1
            :attr:`molecule`.

            This argument can be omitted when the formation energy of a
            :class:`.MacroMolecule` instance is being found, as they
            keep this data stored elsewhere already.

        force_e_calc : :class:`bool`, optional
            If this is ``True`` then all molecules in
            `building_blocks`, `products` and :attr:`molecule` will
            have their energies recalculated. Even if the energy values
            have already been found with the chosen forcefield and
            method. If ``False`` the energy is only calculated if the
            value has not already been found.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        float : :class:`float`
            The pseudoformation energy.

        """

        # Get the function used to calculate the energies.
        efunc = getattr(self, func.name)
        # Get the key of the function used to calculate the energies.
        fkey = func_key(efunc, None, func.params)

        if building_blocks is None:
            building_blocks = ((n, mol) for mol, n in
                               self.molecule.bb_counter.items())

        # Recalculate energies if requested.
        if force_e_calc:
            for _, mol in building_blocks:
                molf = getattr(mol.energy, func.name)
                molf(**func.params)

            getattr(self, func.name)(**func.params)

        # Sum the energy of building blocks under the chosen
        # forcefield.
        e_reactants = 0
        for n, mol in building_blocks:
            # If the calculation has not been done already, perform it.
            if fkey not in mol.energy.values.keys():
                molf = getattr(mol.energy, func.name)
                molf(**func.params)

            e_reactants += n * mol.energy.values[fkey]

        # Get the energy of `self.molecule`. The only product whose
        # energy matters in pseudoformation.
        e_products = (self.values[fkey] if fkey in self.values.keys()
                      else getattr(self, func.name)(**func.params))

        eng = e_reactants - e_products
        return eng

    def rdkit(self, forcefield, conformer=-1):
        """
        Uses ``rdkit`` to calculate the energy.

        Parameters
        ----------
        forcefield : :class:`str`
            The name of the forcefield to be used.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        :class:`float`
            The calculated energy.

        """

        logger.debug('Starting rdkit energy calculation.')
        if forcefield == 'uff':
            self.molecule.mol.UpdatePropertyCache()
            ff = rdkit.UFFGetMoleculeForceField(self.molecule.mol,
                                                confId=conformer)
        if forcefield == 'mmff':
            rdkit.GetSSSR(self.molecule.mol)
            self.molecule.mol.UpdatePropertyCache()
            ff = rdkit.MMFFGetMoleculeForceField(
                  self.molecule.mol,
                  rdkit.MMFFGetMoleculeProperties(self.molecule.mol),
                  confId=conformer)

        eng = ff.CalcEnergy()
        return eng

    @exclude('macromodel_path')
    def macromodel(self, forcefield, macromodel_path, conformer=-1):
        """
        Calculates the energy using macromodel.

        Note that this requires macromodel to be installed and have a
        valid license.

        Parameters
        ----------
        forcefield : :class:`int`
            The id number of the forcefield to be used by macromodel.

        macromodel_path : :class:`str`
            The full path to the Schrodinger suite within the
            user's machine. For example, in a default Linux
            installation the folder will probably be something like
            ``'/opt/schrodinger2017-2'``.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        :class:`float`
            The calculated energy.

        Raises
        ------
        :class:`EnergyError`
            This exception is raised if no energy value is found in the
            MacroModel calculation's ``.log`` file. Likely due to a
            forcefield error.

        """

        # To prevent conflicts when running this function in parallel,
        # a temporary copy of the molecular structure file is made and
        # used for macromodel calculations.

        # Unique file name is generated by inserting a random int into
        # the file path.
        tmp_file = "{}.mol".format(uuid4().int)
        self.molecule.write(tmp_file, conformer)

        file_root, ext = os.path.splitext(tmp_file)
        convrt_app = os.path.join(macromodel_path,
                                  'utilities',
                                  'structconvert')
        convrt_cmd = [convrt_app,
                      tmp_file,
                      file_root+'.mae']
        sp.call(convrt_cmd, stdout=sp.PIPE, stderr=sp.PIPE)

        # Create an input file and run it.
        input_script = (
         "{0}.mae\n"
         "{0}-out.maegz\n"
         " MMOD       0      1      0      0     0.0000     0.0000     "
         "0.0000     0.0000\n"
         " FFLD{1:8}      1      0      0     1.0000     0.0000     "
         "0.0000     0.0000\n"
         " BGIN       0      0      0      0     0.0000     0.0000     "
         "0.0000     0.0000\n"
         " READ      -1      0      0      0     0.0000     0.0000     "
         "0.0000     0.0000\n"
         " ELST      -1      0      0      0     0.0000     0.0000     "
         "0.0000     0.0000\n"
         " WRIT       0      0      0      0     0.0000     0.0000     "
         "0.0000     0.0000\n"
         " END       0      0      0      0     0.0000     0.0000     "
         "0.0000     0.0000\n\n"
        ).format(file_root, forcefield)

        with open(file_root+'.com', 'w') as f:
            f.write(input_script)

        cmd = [os.path.join(macromodel_path, 'bmin'),
               file_root,
               "-WAIT",
               "-LOCAL"]
        sp.call(cmd)

        # Check if the license was found. If not run the function
        # again.
        with open(file_root+'.log', 'r') as f:
            log_content = f.read()
        if ('FATAL -96: Could not check out a license for mmlibs' in
           log_content):
            return self.macromodel(forcefield,
                                   macromodel_path,
                                   conformer)

        # Read the .log file and return the energy.
        with open(file_root+'.log', 'r') as f:
            for line in f:
                if "                   Total Energy =" in line:
                    eng = float(line.split()[-2].replace("=", ""))

        try:
            return eng
        except UnboundLocalError:
            raise EnergyError('MacroModel energy calculation failed.')

    @exclude('mopac_path')
    def mopac(self, mopac_path, settings=None):
        """
        Calculates the energy using MOPAC.

        Note that this requires MOPAC to be installed and have a
        valid license.

        Parameters
        ----------
        settings : :class:`dict`, optional
            A dictionary which maps the names of the optimization
            parameters to their values. Valid values are:

                'hamiltonian' : :class:`str` (default = ``'PM7'``
                    A series of different methods can be selected:
                    PM7, PM6, AM1, CIS (CISD, CISDT), MNDO, RM1, etc..

                    PM7 is the latest version of the reparametrization
                    of NDDO theory, where all the atomic and diatomic
                    parameters were re-optimized / updated from PM6
                    [#]_.

                'eps' : :class:`float` (default = ``80.1``)
                    Sets the dielectric constant for the solvent.
                    Presence of this keyword will cause the COSMO
                    (Conductor-like Screening Model) method to be used
                    to approximate the effect of a solvent model
                    surrounding the molecule. Solvents with a low
                    dielectric constant are not likely to work well
                    with this model. ``0`` means that the dielectric
                    constant is not included in the calculation.
                    ``80.1`` can be used to model a water environment
                    at room temperature.

                'charge' : :class:`float` (default = ``0``)
                    The charge of the system.

                'timeout' : :class:`float` (default = ``172800``)
                    The amount in seconds the calculation is allowed to
                    run before being terminated. The default value is
                    ``2`` days or ``172,800`` seconds.

        mopac_path : :class:`str`
            The full path to the MOPAC installation.

        Returns
        -------
        :class:`float`
            The calculated energy.

        References
        ----------
        .. [#] http://openmopac.net/PM7_accuracy/PM7_accuracy.html

        """

        if settings is None:
            settings = {}

        # Define default vals for the MOPAC input
        vals = {
                'hamiltonian': 'PM7',
                'method': 'NOOPT',
                'eps': 80.1,
                'charge': 0,
                'timeout': 172800,
                }

        vals.update(settings)

        # To prevent conflicts when running this function in parallel,
        # a temporary copy of the molecular structure file is made and
        # used for mopac calculations.

        # Unique file name is generated by inserting a random int into
        # the file path.
        tmp_file = "{}.mol".format(uuid4().int)
        self.molecule.write(tmp_file)

        file_root, ext = os.path.splitext(tmp_file)

        # Generate the input file
        _create_mop(file_root, self.molecule, vals)
        # Run MOPAC
        _run_mopac(file_root, mopac_path)
        return _extract_MOPAC_en(file_root)

    @exclude('mopac_path')
    def mopac_dipole(self, mopac_path, settings=None):
        """
        Calculates the dipole moment using MOPAC.

        Note that this requires MOPAC to be installed and have a
        valid license.

        Parameters
        ----------
        settings : :class:`dict`, optional
            A dictionary which maps the names of the optimization
            parameters to their values. Valid values are:

                'hamiltonian' : :class:`str` (default = ``'PM7'``
                    A series of different methods can be selected:
                    PM7, PM6, AM1, CIS (CISD, CISDT), MNDO, RM1, etc..

                    PM7 is the latest version of the reparametrization
                    of NDDO theory, where all the atomic and diatomic
                    parameters were re-optimized / updated from PM6
                    [#]_.

                'eps' : :class:`float` (default = ``80.1``)
                    Sets the dielectric constant for the solvent.
                    Presence of this keyword will cause the COSMO
                    (Conductor-like Screening Model) method to be used
                    to approximate the effect of a solvent model
                    surrounding the molecule. Solvents with a low
                    dielectric constant are not likely to work well
                    with this model. ``0`` means that the dielectric
                    constant is not included in the calculation.
                    ``80.1`` can be used to model a water environment
                    at room temperature.

                'charge' : :class:`float` (default = ``0``)
                    The charge of the system.

                'timeout' : :class:`float` (default = ``172800``)
                    The amount in seconds the calculation is allowed to
                    run before being terminated. The default value is
                    ``2`` days or ``172,800`` seconds.

        mopac_path : :class:`str`
            The full path to the MOPAC installation.

        Returns
        -------
        :class:`float`
            The calculated dipole.

        References
        ----------
        .. [#] http://openmopac.net/PM7_accuracy/PM7_accuracy.html

        """

        if settings is None:
            settings = {}

        # Define default vals for the MOPAC input
        vals = {
                'hamiltonian': 'PM7',
                'method': 'NOOPT',
                'eps': 80.1,
                'charge': 0,
                'timeout': 172800,
                }

        vals.update(settings)

        # To prevent conflicts when running this function in parallel,
        # a temporary copy of the molecular structure file is made and
        # used for mopac calculations.

        # Unique file name is generated by inserting a random int into
        # the file path.
        tmp_file = "{}.mol".format(uuid4().int)
        self.molecule.write(tmp_file)

        file_root, ext = os.path.splitext(tmp_file)

        # Generate the input file
        _create_mop(file_root, self.molecule, vals)
        # Run MOPAC
        _run_mopac(file_root, mopac_path)
        return _extract_MOPAC_dipole(file_root)

    @exclude('mopac_path')
    def mopac_ea(self, mopac_path, settings=None):
        """
        Calculates the electron affinity using MOPAC.

        Note that this requires MOPAC to be installed and have a
        valid license.

        Parameters
        ----------
        settings : :class:`dict`, optional
            A dictionary which maps the names of the optimization
            parameters to their values. Valid values are:

                'hamiltonian' : :class:`str` (default = ``'PM7'``
                    A series of different methods can be selected:
                    PM7, PM6, AM1, CIS (CISD, CISDT), MNDO, RM1, etc..

                    PM7 is the latest version of the reparametrization
                    of NDDO theory, where all the atomic and diatomic
                    parameters were re-optimized / updated from PM6
                    [#]_.

                'eps' : :class:`float` (default = ``80.1``)
                    Sets the dielectric constant for the solvent.
                    Presence of this keyword will cause the COSMO
                    (Conductor-like Screening Model) method to be used
                    to approximate the effect of a solvent model
                    surrounding the molecule. Solvents with a low
                    dielectric constant are not likely to work well
                    with this model. ``0`` means that the dielectric
                    constant is not included in the calculation.
                    ``80.1`` can be used to model a water environment
                    at room temperature.

                'charge' : :class:`float` (default = ``0``)
                    The charge of the system.

                'timeout' : :class:`float` (default = ``172800``)
                    The amount in seconds the calculation is allowed to
                    run before being terminated. The default value is
                    ``2`` days or ``172,800`` seconds.

        mopac_path : :class:`str`
            The full path to the MOPAC installation.

        Returns
        -------
        :class:`float`
            The calculated energy.

        References
        ----------
        .. [#] http://openmopac.net/PM7_accuracy/PM7_accuracy.html

        """

        if settings is None:
            settings = {}

        # Define default vals for the MOPAC input
        vals = {
                'hamiltonian': 'PM7',
                'method': 'NOOPT',
                'eps': 80.1,
                'charge': 0,
                'timeout': 172800,
                }
        vals.update(settings)

        # First check the energy of the neutral system
        # To prevent conflicts when running this function in parallel,
        # a temporary copy of the molecular structure file is made and
        # used for mopac calculations.

        # Unique file name is generated by inserting a random int into
        # the file path.
        tmp_file = "{}.mol".format(uuid4().int)
        self.molecule.write(tmp_file)

        file_root, ext = os.path.splitext(tmp_file)

        # Generate the input file
        _create_mop(file_root, self.molecule, vals)
        # Run MOPAC
        _run_mopac(file_root, mopac_path)

        # Extract the neutral energy
        en1 = _extract_MOPAC_en(file_root)

        # Update the settings for the anion optimization
        settings2 = {
                    'method': 'OPT',
                    'gradient': 0.01,
                    'charge': -1,
                    'fileout': 'PDBOUT'
                    }

        vals.update(settings2)

        # Now generate a new molecule
        mol2 = copy.deepcopy(self.molecule)
        # Run the mopac optimisation
        mopac_opt(mol2, mopac_path, vals)
        # Extract the energy by using the self.mopac method
        vals['method'] = 'NOOPT'
        del vals['gradient']
        del vals['fileout']
        en2 = mol2.energy.mopac(mopac_path, vals)
        # Calculate the EA (eV)
        return en2 - en1

    @exclude('mopac_path')
    def mopac_ip(self, mopac_path, settings=None):
        """
        Calculates the ionization potential using MOPAC.

        Note that this requires MOPAC to be installed and have a
        valid license.

        Parameters
        ----------
        settings : :class:`dict`, optional
            A dictionary which maps the names of the optimization
            parameters to their values. Valid values are:

                'hamiltonian' : :class:`str` (default = ``'PM7'``
                    A series of different methods can be selected:
                    PM7, PM6, AM1, CIS (CISD, CISDT), MNDO, RM1, etc..

                    PM7 is the latest version of the reparametrization
                    of NDDO theory, where all the atomic and diatomic
                    parameters were re-optimized / updated from PM6
                    [#]_.

                'eps' : :class:`float` (default = ``80.1``)
                    Sets the dielectric constant for the solvent.
                    Presence of this keyword will cause the COSMO
                    (Conductor-like Screening Model) method to be used
                    to approximate the effect of a solvent model
                    surrounding the molecule. Solvents with a low
                    dielectric constant are not likely to work well
                    with this model. ``0`` means that the dielectric
                    constant is not included in the calculation.
                    ``80.1`` can be used to model a water environment
                    at room temperature.

                'charge' : :class:`float` (default = ``0``)
                    The charge of the system.

                'timeout' : :class:`float` (default = ``172800``)
                    The amount in seconds the calculation is allowed to
                    run before being terminated. The default value is
                    ``2`` days or ``172,800`` seconds.

        mopac_path : :class:`str`
            The full path to the MOPAC installation.

        Returns
        -------
        :class:`float`
            The calculated energy.

        References
        ----------
        .. [#] http://openmopac.net/PM7_accuracy/PM7_accuracy.html

        """

        if settings is None:
            settings = {}

        # Define default vals for the MOPAC input
        vals = {
                'hamiltonian': 'PM7',
                'method': 'NOOPT',
                'eps': 80.1,
                'charge': 0,
                'timeout': 172800,
                }
        vals.update(settings)

        # First check the energy of the neutral system
        # To prevent conflicts when running this function in parallel,
        # a temporary copy of the molecular structure file is made and
        # used for mopac calculations.

        # Unique file name is generated by inserting a random int into
        # the file path.
        tmp_file = "{}.mol".format(uuid4().int)
        self.molecule.write(tmp_file)

        file_root, ext = os.path.splitext(tmp_file)

        # Generate the input file
        _create_mop(file_root, self.molecule, vals)
        # Run MOPAC
        _run_mopac(file_root, mopac_path)

        # Extract the neutral energy
        en1 = _extract_MOPAC_en(file_root)

        # Update the settings for the cation optimization
        settings2 = {
                    'method': 'OPT',
                    'gradient': 0.01,
                    'charge': 1,
                    'fileout': 'PDBOUT'
                    }

        vals.update(settings2)

        # Now generate a new molecule
        mol2 = copy.deepcopy(self.molecule)
        # Run the mopac optimisation
        mopac_opt(mol2, mopac_path, vals)
        # Extract the energy by using the self.mopac method
        vals['method'] = 'NOOPT'
        del vals['gradient']
        del vals['fileout']
        en2 = mol2.energy.mopac(mopac_path, vals)
        # Calculate the IP (eV)
        return en2 - en1


def formation_key(fargs, fkwargs):
    """
    Generates the key of :meth:`Energy.formation`.

    Parameters
    ----------
    fargs : :class:`tuple`
        The arguments with which :meth:`Energy.formation` was called.

    fkwargs : :class:`dict`
        The keyword arguments with which :meth:`Energy.formation` was
        called.

    Returns
    -------
    :class:`.FunctionData`
        The :class:`.FunctionData` object representing the key of
        :meth:`Energy.formation`, when called with the arguments
        `fargs` and keyword arguments `fkwargs`.

    """

    fsig = sig(Energy.formation)

    # Get a dictionary of all the supplied parameters.
    bound = dict(fsig.bind_partial(*fargs, **fkwargs).arguments)
    # Get a dictionary of all the default initialized parameters.
    default = {key: value.default for key, value in
               dict(fsig.parameters).items() if
               key not in bound.keys()}

    # Combine the two sets of parameters and get rid of the `self`
    # parameter, if present.
    bound.update(default)
    if 'self' in bound.keys():
        bound.pop('self')

    # Replace the energy function to be used with the key of the
    # energy function to be used.
    efuncdata = bound['func']
    efunc = getattr(Energy, efuncdata.name)
    bound['func'] = func_key(efunc, None, efuncdata.params)

    # Don't want this paramter in the key as it doenst affect the
    # result.
    bound.pop('force_e_calc')

    # Return an FunctionData object representing the function and
    # chosen parameters.
    return FunctionData('formation', **bound)


Energy.formation.key = formation_key


def pseudoformation_key(fargs, fkwargs):
    """
    Generates key of the :meth:`Energy.pseudoformation`.

    Parameters
    ----------
    fargs : :class:`tuple`
        The arguments with which :meth:`Energy.pseudoformation` was
        called.

    fkwargs : :class:`dict`
        The keyword arguments with which :meth:`Energy.pseudoformation`
        was called.

    Returns
    -------
    :class:`.FunctionData`
        The :class:`.FunctionData` object representing the key when
        :meth:`Energy.pseudoformation` is called with the arguments
        `fargs` and keyword arguments `fkwargs`.

    """

    fsig = sig(Energy.pseudoformation)

    # Get a dictionary of all the supplied parameters.
    bound = dict(fsig.bind_partial(*fargs, **fkwargs).arguments)
    # Get a dictionary of all the default initialized parameters.
    default = {key: value.default for key, value in
               dict(fsig.parameters).items() if
               key not in bound.keys()}

    # Combine the two sets of parameters and get rid of the `self`
    # parameter, if present.
    bound.update(default)
    if 'self' in bound.keys():
        bound.pop('self')

    # Replace the energy function to be used with the key of the
    # energy function to be used.
    efuncdata = bound['func']
    efunc = getattr(Energy, efuncdata.name)
    bound['func'] = func_key(efunc, None, efuncdata.params)

    # Don't want this paramter in the key as it doenst affect the
    # result.
    bound.pop('force_e_calc')

    # Return an FunctionData object representing the function and
    # chosen parameters.
    return FunctionData('pseudoformation', **bound)


Energy.pseudoformation.key = pseudoformation_key


def _run_mopac(file_root, mopac_path, timeout=3600):

    mop_file = file_root + '.mop'

    logger.info(f'Running MOPAC - {file_root}.')

    # To run MOPAC a command is issued to the console via
    # ``subprocess.Popen``. The command is the full path of the
    # ``mopac`` program.
    file_root, ext = os.path.splitext(mop_file)
    opt_cmd = [mopac_path, file_root]
    opt_proc = psutil.Popen(opt_cmd, stdout=sp.PIPE,
                            stderr=sp.STDOUT,
                            universal_newlines=True)

    try:
        if timeout:
            proc_out, _ = opt_proc.communicate(timeout=timeout)
        else:
            proc_out, _ = opt_proc.communicate()
    except sp.TimeoutExpired:
        logger.info(('Minimization took too long and was terminated '
                     'by force - {}').format(file_root))
        _kill_mopac(file_root)


def _kill_mopac(file_root):
    """
    Stops an in-progress MOPAC run.

    To kill a MOPAC run, a file with the molecule's name and a ``.end``
    extension is written.

    Parameters
    ----------
    file_root : :class:`str`
        The molecule's name.

    Returns
    -------
    None : :class:`NoneType`

    """
    end_file = file_root + '.end'

    with open(end_file, 'w') as end:
        end.write('SHUT')


def _mop_line(settings):
    """
    Formats `settings` into a MOPAC input string.

    Parameters
    ----------
    settings : :class:`dict`
        Dictionary defined in :func:`mopac_opt`, where all the run
        details are defined.

    Returns
    -------
    :class:`str`
        String containing all the MOPAC keywords correctly formatted
        for the input file.

    """

    # Generate an empty string
    mopac_run_str = ""

    # Add Hamiltonian info
    mopac_run_str = mopac_run_str + settings['hamiltonian']
    # Forcing a single point calculation
    mopac_run_str = mopac_run_str + ' NOOPT '
    # Add EPS info
    eps_info = ' EPS={} '.format(settings['eps'])
    mopac_run_str = mopac_run_str + eps_info
    # Add Charge info
    charge_info = ' CHARGE={} '.format(settings['charge'])
    mopac_run_str = mopac_run_str + charge_info

    return mopac_run_str


def _create_mop(file_root, molecule, settings):
    """
    Creates the ``.mop`` file holding the molecule to be optimized.

    Parameters
    ----------
    mol : :class:`.Molecule`
        The molecule which is to be optimized. Its molecular
        structure file is converted to a ``.mop`` file. The original
        file is also kept.

    settings : :class:`dict`
        Dictionary defined by the MOPAC methods, where all the run
        details are defined.

    Returns
    -------
    :class:`str`
        The full path of the newly created ``.mop`` file.

    """

    mop_file = file_root + '.mop'
    mol = molecule.mol

    logger.info('Creating .mop file - {}.'.format(file_root))

    # Generate the mop file containing the MOPAC run info
    with open(mop_file, 'w') as mop:
        # line for the run info
        mop.write(_mop_line(settings) + "\n")
        # line with the name of the molecule
        mop.write(file_root + "\n\n")

        # print the structural info
        for atom in mol.GetAtoms():
            atom_id = atom.GetIdx()
            atom_symbol = atom.GetSymbol()
            x, y, z = mol.GetConformer().GetAtomPosition(atom_id)
            atom_info = "{}   {}   +1  {}   +1  {}   +1 \n".format(atom_symbol,
                                                                   x, y, z)
            mop.write(atom_info)

    return mop_file


def _extract_MOPAC_en(file_root):
    mopac_out = file_root + '.arc'

    with open(mopac_out) as outfile:
        target = "TOTAL ENERGY"
        energy_str = str([x for x in outfile.readlines() if target in x][0])
        energy_val = float(energy_str.split()[3])

    return energy_val


def _extract_MOPAC_dipole(file_root):
    mopac_out = file_root + '.arc'

    with open(mopac_out) as outfile:
        target = "DIPOLE"
        energy_str = str([x for x in outfile.readlines() if target in x][0])
        energy_val = float(energy_str.split()[2])

    return energy_val
